"""Date parsing utilities."""
import time
from calendar import monthrange
from datetime import datetime, timezone
from typing import Dict, Optional, Tuple

import dateutil
from dateutil.parser import ParserError, _timelex, parserinfo

default_sd_year = 1
default_date_notz = datetime(
    year=default_sd_year,
    month=1,
    day=1,
    hour=0,
    minute=0,
    second=0,
    microsecond=0,
)
default_date = datetime(
    year=default_sd_year,
    month=1,
    day=1,
    hour=0,
    minute=0,
    second=0,
    microsecond=0,
    tzinfo=timezone.utc,
)
default_ed_year = 9990
default_enddate_notz = datetime(
    year=default_ed_year,
    month=12,
    day=31,
    hour=0,
    minute=0,
    second=0,
    microsecond=0,
)
default_enddate = datetime(
    year=default_ed_year,
    month=12,
    day=31,
    hour=0,
    minute=0,
    second=0,
    microsecond=0,
    tzinfo=timezone.utc,
)

# Used https://stackoverflow.com/questions/1703546/parsing-date-time-string-with-timezone-abbreviated-name-in-python/4766400#4766400
# with IST added
default_timezone_info = """-12 Y
-11 X NUT SST
-10 W CKT HAST HST TAHT TKT
-9 V AKST GAMT GIT HADT HNY
-8 U AKDT CIST HAY HNP PST PT
-7 T HAP HNR MST PDT
-6 S CST EAST GALT HAR HNC MDT
-5 R CDT COT EASST ECT EST ET HAC HNE PET
-4 Q AST BOT CLT COST EDT FKT GYT HAE HNA PYT
-3 P ADT ART BRT CLST FKST GFT HAA PMST PYST SRT UYT WGT
-2 O BRST FNT PMDT UYST WGST
-1 N AZOT CVT EGT
0 Z EGST GMT UTC WET WT
1 A CET DFT WAT WEDT WEST
2 B CAT CEDT CEST EET SAST WAST
3 C EAT EEDT EEST IDT MSK
4 D AMT AZT GET GST KUYT MSD MUT RET SAMT SCT
5 E AMST AQTT AZST HMT MAWT MVT PKT TFT TJT TMT UZT YEKT
6 F ALMT BIOT BTT IOT KGT NOVT OMST YEKST
7 G CXT DAVT HOVT ICT KRAT NOVST OMSST THA WIB
8 H ACT AWST BDT BNT CAST HKT IRKT KRAST MYT PHT SGT ULAT WITA WST
9 I AWDT IRKST JST KST PWT TLT WDT WIT YAKT
10 K AEST ChST PGT VLAT YAKST YAPT
11 L AEDT LHDT MAGT NCT PONT SBT VLAST VUT
12 M ANAST ANAT FJT GILT MAGST MHT NZST PETST PETT TVT WFT
13 FJST NZDT
11.5 NFT
10.5 ACDT LHST
9.5 ACST
6.5 CCT MMT
5.75 NPT
5.5 IST SLT
4.5 AFT IRDT
3.5 IRST
-2.5 HAT NDT
-3.5 HNT NST NT
-4.5 HLV VET
-9.5 MART MIT"""


def get_tzinfos(timezone_info: str) -> Dict[str, int]:
    """Get tzinfos dictionary used by dateutil from timezone information
    string.

    Args:
        timezone_info (str): Timezones information string

    Returns:
        Dict[str, int]: tzinfos dictionary
    """
    tzinfos = {}
    for tz_descr in map(str.split, timezone_info.split("\n")):
        tz_offset = int(float(tz_descr[0]) * 3600)
        for tz_code in tz_descr[1:]:
            tzinfos[tz_code] = tz_offset
    return tzinfos


default_tzinfos = get_tzinfos(default_timezone_info)


# Ugly copy and paste from dateutil.parser._parser._ymd with dayfirst modified to mean day is outer value
# ie. dayfirst prefers dmy or ymd where in dateutil it prefers dmy and ydm!
class _ymd(list):  # pragma: no cover
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        self.century_specified = False
        self.dstridx = None
        self.mstridx = None
        self.ystridx = None

    @property
    def has_year(self):
        return self.ystridx is not None

    @property
    def has_month(self):
        return self.mstridx is not None

    @property
    def has_day(self):
        return self.dstridx is not None

    def could_be_day(self, value):
        if self.has_day:
            return False
        elif not self.has_month:
            return 1 <= value <= 31
        elif not self.has_year:
            # Be permissive, assume leap year
            month = self[self.mstridx]
            return 1 <= value <= monthrange(2000, month)[1]
        else:
            month = self[self.mstridx]
            year = self[self.ystridx]
            return 1 <= value <= monthrange(year, month)[1]

    def append(self, val, label=None):
        if hasattr(val, "__len__"):
            if val.isdigit() and len(val) > 2:
                self.century_specified = True
                if label not in [None, "Y"]:  # pragma: no cover
                    raise ValueError(label)
                label = "Y"
        elif val > 100:
            self.century_specified = True
            if label not in [None, "Y"]:  # pragma: no cover
                raise ValueError(label)
            label = "Y"

        super(self.__class__, self).append(int(val))

        if label == "M":
            if self.has_month:
                raise ValueError("Month is already set")
            self.mstridx = len(self) - 1
        elif label == "D":
            if self.has_day:
                raise ValueError("Day is already set")
            self.dstridx = len(self) - 1
        elif label == "Y":
            if self.has_year:
                raise ValueError("Year is already set")
            self.ystridx = len(self) - 1

    def _resolve_from_stridxs(self, strids):
        """Try to resolve the identities of year/month/day elements using
        ystridx, mstridx, and dstridx, if enough of these are specified."""
        if len(self) == 3 and len(strids) == 2:
            # we can back out the remaining stridx value
            missing = [x for x in range(3) if x not in strids.values()]
            key = [x for x in ["y", "m", "d"] if x not in strids]
            assert len(missing) == len(key) == 1
            key = key[0]
            val = missing[0]
            strids[key] = val

        assert len(self) == len(strids)  # otherwise this should not be called
        out = {key: self[strids[key]] for key in strids}
        return (out.get("y"), out.get("m"), out.get("d"))

    def resolve_ymd(self, yearfirst, dayfirst):
        len_ymd = len(self)
        year, month, day = (None, None, None)

        strids = (
            ("y", self.ystridx),
            ("m", self.mstridx),
            ("d", self.dstridx),
        )

        strids = {key: val for key, val in strids if val is not None}
        if len(self) == len(strids) > 0 or (
            len(self) == 3 and len(strids) == 2
        ):
            return self._resolve_from_stridxs(strids)

        mstridx = self.mstridx

        if len_ymd > 3:
            raise ValueError("More than three YMD values")
        elif len_ymd == 1 or (mstridx is not None and len_ymd == 2):
            # One member, or two members with a month string
            if mstridx is not None:
                month = self[mstridx]
                # since mstridx is 0 or 1, self[mstridx-1] always
                # looks up the other element
                other = self[mstridx - 1]
            else:
                other = self[0]

            if len_ymd > 1 or mstridx is None:
                if other > 31:
                    year = other
                else:
                    day = other

        elif len_ymd == 2:
            # Two members with numbers
            if self[0] > 31:
                # 99-01
                year, month = self
            elif self[1] > 31:
                # 01-99
                month, year = self
            elif dayfirst and self[1] <= 12:
                # 13-01
                day, month = self
            else:
                # 01-13
                month, day = self

        elif len_ymd == 3:
            # Three members
            if mstridx == 0:
                if self[1] > 31:
                    # Apr-2003-25
                    month, year, day = self
                else:
                    month, day, year = self
            elif mstridx == 1:
                if self[0] > 31 or (yearfirst and self[2] <= 31):
                    # 99-Jan-01
                    year, month, day = self
                else:
                    # 01-Jan-01
                    # Give precedence to day-first, since
                    # two-digit years is usually hand-written.
                    day, month, year = self

            elif mstridx == 2:
                # WTF!?
                if self[1] > 31:
                    # 01-99-Jan
                    day, year, month = self
                else:
                    # 99-01-Jan
                    year, day, month = self

            else:
                if (
                    self[0] > 31
                    or self.ystridx == 0
                    or (yearfirst and self[1] <= 12 and self[2] <= 31)
                ):
                    # 99-01-01
                    if dayfirst and self[1] <= 12:  # CHANGED
                        year, month, day = self
                    else:
                        year, day, month = self
                elif self[0] > 12 or (dayfirst and self[1] <= 12):
                    # 13-01-01
                    day, month, year = self
                else:
                    # 01-13-01
                    month, day, year = self

        return year, month, day


# Ugly copy and paste from dateutil.dateparser with minor changes
class DateParser(dateutil.parser.parser):  # pragma: no cover
    def _parse(
        self,
        timestr,
        dayfirst=None,
        yearfirst=None,
        fuzzy=False,
        fuzzy_with_tokens=False,
    ):
        """Private method which performs the heavy lifting of parsing, called
        from ``parse()``, which passes on its ``kwargs`` to this function.

        :param timestr:
            The string to parse.

        :param dayfirst:
            Whether to interpret the first value in an ambiguous 3-integer date
            (e.g. 01/05/09) as the day (``True``) or month (``False``). If
            ``yearfirst`` is set to ``True``, this distinguishes between YDM
            and YMD. If set to ``None``, this value is retrieved from the
            current :class:`parserinfo` object (which itself defaults to
            ``False``).

        :param yearfirst:
            Whether to interpret the first value in an ambiguous 3-integer date
            (e.g. 01/05/09) as the year. If ``True``, the first number is taken
            to be the year, otherwise the last number is taken to be the year.
            If this is set to ``None``, the value is retrieved from the current
            :class:`parserinfo` object (which itself defaults to ``False``).

        :param fuzzy:
            Whether to allow fuzzy parsing, allowing for string like "Today is
            January 1, 2047 at 8:21:00AM".

        :param fuzzy_with_tokens:
            If ``True``, ``fuzzy`` is automatically set to True, and the parser
            will return a tuple where the first element is the parsed
            :class:`datetime.datetime` datetimestamp and the second element is
            a tuple containing the portions of the string which were ignored:

            .. doctest::

                >>> from dateutil.parser import parse
                >>> parse("Today is January 1, 2047 at 8:21:00AM", fuzzy_with_tokens=True)
                (datetime.datetime(2047, 1, 1, 8, 21), (u'Today is ', u' ', u'at '))
        """
        if fuzzy_with_tokens:
            fuzzy = True

        info = self.info

        if dayfirst is None:
            dayfirst = info.dayfirst

        if yearfirst is None:
            yearfirst = info.yearfirst

        res = self._result()
        l = _timelex.split(timestr)  # noqa: E741

        skipped_idxs = []

        # year/month/day list
        ymd = _ymd()

        len_l = len(l)
        i = 0
        try:
            while i < len_l:
                # Check if it's a number
                value_repr = l[i]
                try:
                    value = float(value_repr)
                except ValueError:
                    value = None

                if value is not None:
                    # Numeric token
                    i = self._parse_numeric_token(l, i, info, ymd, res, fuzzy)

                # Check weekday
                elif info.weekday(l[i]) is not None:
                    value = info.weekday(l[i])
                    res.weekday = value

                # Check month name
                elif info.month(l[i]) is not None:
                    value = info.month(l[i])
                    ymd.append(value, "M")

                    if i + 1 < len_l:
                        if l[i + 1] in ("-", "/"):
                            # Jan-01[-99]
                            sep = l[i + 1]
                            ymd.append(l[i + 2])

                            if i + 3 < len_l and l[i + 3] == sep:
                                # Jan-01-99
                                ymd.append(l[i + 4])
                                i += 2

                            i += 2

                        elif (
                            i + 4 < len_l
                            and l[i + 1] == l[i + 3] == " "
                            and info.pertain(l[i + 2])
                        ):
                            # Jan of 01
                            # In this case, 01 is clearly year
                            if l[i + 4].isdigit():
                                # Convert it here to become unambiguous
                                value = int(l[i + 4])
                                year = str(info.convertyear(value))
                                ymd.append(year, "Y")
                            else:
                                # Wrong guess
                                pass
                                # TODO: not hit in tests
                            i += 4

                # Check am/pm
                elif info.ampm(l[i]) is not None:
                    value = info.ampm(l[i])
                    val_is_ampm = self._ampm_valid(res.hour, res.ampm, fuzzy)

                    if val_is_ampm:
                        res.hour = self._adjust_ampm(res.hour, value)
                        res.ampm = value

                    elif fuzzy:
                        skipped_idxs.append(i)

                # Check for a timezone name
                elif self._could_be_tzname(
                    res.hour, res.tzname, res.tzoffset, l[i]
                ):
                    res.tzname = l[i]
                    res.tzoffset = info.tzoffset(res.tzname)

                    # Check for something like GMT+3, or BRST+3. Notice
                    # that it doesn't mean "I am 3 hours after GMT", but
                    # "my time +3 is GMT". If found, we reverse the
                    # logic so that timezone parsing code will get it
                    # right.
                    if i + 1 < len_l and l[i + 1] in ("+", "-"):
                        l[i + 1] = ("+", "-")[l[i + 1] == "+"]
                        res.tzoffset = None
                        if info.utczone(res.tzname):
                            # With something like GMT+3, the timezone
                            # is *not* GMT.
                            res.tzname = None

                # Check for a numbered timezone
                elif res.hour is not None and l[i] in ("+", "-"):
                    signal = (-1, 1)[l[i] == "+"]
                    len_li = len(l[i + 1])

                    # TODO: check that l[i + 1] is integer?
                    if len_li == 4:
                        # -0300
                        hour_offset = int(l[i + 1][:2])
                        min_offset = int(l[i + 1][2:])
                    elif i + 2 < len_l and l[i + 2] == ":":
                        # -03:00
                        hour_offset = int(l[i + 1])
                        min_offset = int(
                            l[i + 3]
                        )  # TODO: Check that l[i+3] is minute-like?
                        i += 2
                    elif len_li <= 2:
                        # -[0]3
                        hour_offset = int(l[i + 1][:2])
                        min_offset = 0
                    else:
                        raise ValueError(timestr)

                    res.tzoffset = signal * (
                        hour_offset * 3600 + min_offset * 60
                    )

                    # Look for a timezone name between parenthesis
                    if (
                        i + 5 < len_l
                        and info.jump(l[i + 2])
                        and l[i + 3] == "("
                        and l[i + 5] == ")"
                        and 3 <= len(l[i + 4])
                        and self._could_be_tzname(
                            res.hour, res.tzname, None, l[i + 4]
                        )
                    ):
                        # -0300 (BRST)
                        res.tzname = l[i + 4]
                        i += 4

                    i += 1

                # Check jumps
                elif not (info.jump(l[i]) or fuzzy):
                    raise ValueError(timestr)

                else:
                    skipped_idxs.append(i)
                i += 1

            # Process year/month/day
            year, month, day = ymd.resolve_ymd(yearfirst, dayfirst)

            res.century_specified = ymd.century_specified
            res.year = year
            res.month = month
            res.day = day
        # *** CHANGES FROM HERE DOWN TO 3 VALUES IN TUPLE ***
        except (IndexError, ValueError):
            return None, None, None

        if not info.validate(res):
            return None, None, None

        if fuzzy_with_tokens:
            skipped_tokens, date_tokens = self._recombine_skipped_date(
                l, skipped_idxs
            )
            return res, tuple(skipped_tokens), tuple(date_tokens)
        else:
            return res, None, None

    def _recombine_skipped_date(self, tokens, skipped_idxs):
        """
        >>> tokens = ["foo", " ", "bar", " ", "19June2000", "baz"]
        >>> skipped_idxs = [0, 1, 2, 5]
        >>> _recombine_skipped_date(tokens, skipped_idxs)
        ["foo bar", "baz"], ["19June2000"]
        """
        skipped_tokens = []
        date_tokens = []
        prev = None
        for idx, token in enumerate(tokens):
            if idx in skipped_idxs:
                if prev is None or prev == "date":
                    skipped_tokens.append(token)
                else:
                    skipped_tokens[-1] = skipped_tokens[-1] + token
                prev = "skipped"
            else:
                if prev is None or prev == "skipped":
                    date_tokens.append(token)
                else:
                    date_tokens[-1] = date_tokens[-1] + token
                prev = "date"

        return skipped_tokens, date_tokens


DEFAULTPARSER = DateParser(parserinfo(dayfirst=True))


def parse(
    timestr, default=None, ignoretz=False, tzinfos=None, **kwargs
):  # pragma: no cover
    """Parse the date/time string into a :class:`datetime.datetime` object.

    :param timestr:
        Any date/time string using the supported formats.

    :param default:
        The default datetime object, if this is a datetime object and not
        ``None``, elements specified in ``timestr`` replace elements in the
        default object.

    :param ignoretz:
        If set ``True``, time zones in parsed strings are ignored and a
        naive :class:`datetime.datetime` object is returned.

    :param tzinfos:
        Additional time zone names / aliases which may be present in the
        string. This argument maps time zone names (and optionally offsets
        from those time zones) to time zones. This parameter can be a
        dictionary with timezone aliases mapping time zone names to time
        zones or a function taking two parameters (``tzname`` and
        ``tzoffset``) and returning a time zone.

        The timezones to which the names are mapped can be an integer
        offset from UTC in seconds or a :class:`tzinfo` object.

        .. doctest::
           :options: +NORMALIZE_WHITESPACE

            >>> from dateutil.parser import parse
            >>> from dateutil.tz import gettz
            >>> tzinfos = {"BRST": -7200, "CST": gettz("America/Chicago")}
            >>> parse("2012-01-19 17:21:00 BRST", tzinfos=tzinfos)
            datetime.datetime(2012, 1, 19, 17, 21, tzinfo=tzoffset(u'BRST', -7200))
            >>> parse("2012-01-19 17:21:00 CST", tzinfos=tzinfos)
            datetime.datetime(2012, 1, 19, 17, 21,
                              tzinfo=tzfile('/usr/share/zoneinfo/America/Chicago'))

        This parameter is ignored if ``ignoretz`` is set.

    :param \\*\\*kwargs:
        Keyword arguments as passed to ``_parse()``.

    :return:
        Returns a :class:`datetime.datetime` object or, if the
        ``fuzzy_with_tokens`` option is ``True``, returns a tuple, the
        first element being a :class:`datetime.datetime` object, the second
        a tuple containing the fuzzy tokens.

    :raises ParserError:
        Raised for invalid or unknown string format, if the provided
        :class:`tzinfo` is not in a valid format, or if an invalid date
        would be created.

    :raises TypeError:
        Raised for non-string or character stream input.

    :raises OverflowError:
        Raised if the parsed date exceeds the largest valid C integer on
        your system.
    """

    if default is None:
        default = datetime.now().replace(
            hour=0, minute=0, second=0, microsecond=0
        )

    res, skipped_tokens, date_tokens = DEFAULTPARSER._parse(timestr, **kwargs)

    if res is None:
        raise ParserError("Unknown string format: %s", timestr)

    if len(res) == 0:
        raise ParserError("String does not contain a date: %s", timestr)

    try:
        ret = DEFAULTPARSER._build_naive(res, default)
    except ValueError as e:
        raise ParserError(str(e) + f": {timestr}") from e

    if not ignoretz:
        ret = DEFAULTPARSER._build_tzaware(ret, res, tzinfos)

    if kwargs.get("fuzzy_with_tokens", False):
        return ret, skipped_tokens, date_tokens
    else:
        return ret


def now_utc() -> datetime:
    """Return now with UTC timezone.

    Returns:
        datetime: Now with UTC timezone
    """
    return datetime.now(timezone.utc)


def parse_date_range(
    string: str,
    date_format: Optional[str] = None,
    timezone_handling: int = 0,
    fuzzy: Optional[Dict] = None,
    include_microseconds: bool = False,
    zero_time: bool = False,
    max_starttime: bool = False,
    max_endtime: bool = False,
    default_timezones: Optional[str] = None,
) -> Tuple[datetime, datetime]:
    """Parse date from string using specified date_format if given and return
    datetime date range in dictionary keys startdate and enddate. If no
    date_format is supplied, the function will guess, which for unambiguous
    formats, should work fine.

    By default, no timezone information will be parsed and the returned datetime will
    have timezone UTC. To change this behaviour, timezone_handling should be changed
    from its default of 0. If it is 1, then no timezone information will be parsed and a
    naive datetime will be returned. If it is 2 or more, then timezone information will
    be parsed. For 2, failure to parse timezone will result in a naive datetime. For 3,
    failure to parse timezone will result in the timezone being set to UTC. For 4 and 5,
    the time will be converted from whatever timezone is identified to UTC. For 4,
    failure to parse timezone will result in a naive (local) datetime converted to UTC.
    For 5, failure to parse timezone will result in the timezone being set to UTC.

    To parse a date within a string containing other text, you can supply a dictionary
    in the fuzzy parameter. In this case, dateutil's fuzzy parsing is used and the
    results returned in the dictionary in keys startdate, enddate, date (the string
    elements used to make the date) and nondate (the non date part of the string).

    By default, microseconds are ignored (set to 0), but can be included by setting
    include_microseconds to True. Any time elements are set to 0 if zero_time is True.
    If max_starttime is True, then the start date's time is set to 23:59:59. If
    max_endtime is True, then the end date's time is set to 23:59:59.

    When inferring time zones, a default set of time zones will be used unless
    overridden by passing in default_timezones which is a string of the form:

        -11 X NUT SST
        -10 W CKT HAST HST TAHT TKT

    Args:
        string (str): Dataset date string
        date_format (Optional[str]): Date format. If None is given, will attempt to guess. Defaults to None.
        timezone_handling (int): Timezone handling. See description. Defaults to 0 (ignore timezone, return UTC).
        fuzzy (Optional[Dict]): If dict supplied, fuzzy matching will be used and results returned in dict
        include_microseconds (bool): Includes microseconds if True. Defaults to False.
        zero_time (bool): Zero time elements of datetime if True. Defaults to False.
        max_starttime (bool): Make start date time component 23:59:59:999999. Defaults to False.
        max_endtime (bool): Make end date time component 23:59:59:999999. Defaults to False.
        default_timezones (Optional[str]): Timezone information. Defaults to None. (Internal default).

    Returns:
        Tuple[datetime,datetime]: Tuple containing start date and end date
    """
    if date_format is None or fuzzy is not None:
        if timezone_handling >= 2:
            if default_timezones is None:
                tzinfos = default_tzinfos
            else:
                tzinfos = get_tzinfos(default_timezones)
            ignoretz = False
        else:
            ignoretz = True
            tzinfos = None

        if fuzzy is not None:
            parsed_string1 = parse(
                string,
                fuzzy_with_tokens=True,
                default=default_date_notz,
                ignoretz=ignoretz,
                tzinfos=tzinfos,
            )
            parsed_string2 = parse(
                string,
                fuzzy_with_tokens=True,
                default=default_enddate_notz,
                ignoretz=ignoretz,
                tzinfos=tzinfos,
            )
            startdate = parsed_string1[0]
            enddate = parsed_string2[0]
            nondate = parsed_string1[1]
            if nondate:
                fuzzy["nondate"] = nondate
            else:
                fuzzy["nondate"] = None
            fuzzy["date"] = parsed_string1[2]
        else:
            startdate = parse(
                string,
                default=default_date_notz,
                ignoretz=ignoretz,
                tzinfos=tzinfos,
            )
            enddate = parse(
                string,
                default=default_enddate_notz,
                ignoretz=ignoretz,
                tzinfos=tzinfos,
            )
        if (
            startdate.year == default_sd_year
            and enddate.year == default_ed_year
        ):
            raise ParserError("No year in date!")
    else:
        try:
            startdate = datetime.strptime(string, date_format)
        except ValueError as e:
            raise ParserError(str(e)) from e
        if (
            startdate.year == 1900 and "%Y" not in date_format
        ):  # 1900 is default when no year supplied
            raise ParserError("No year in date!")
        enddate = startdate
        if not any(x in date_format for x in ["%d", "%j"]):
            startdate = startdate.replace(day=default_date.day)
            endday = default_enddate.day
            not_set = True
            while not_set:
                try:
                    enddate = enddate.replace(day=endday)
                    not_set = False
                except ValueError as e:
                    endday -= 1
                    if endday == 0:
                        raise ParserError(
                            f"No end day of month found for {str(enddate)}!"
                        ) from e
        if not any(str in date_format for str in ["%b", "%B", "%m", "%j"]):
            startdate = startdate.replace(month=default_date.month)
            enddate = enddate.replace(month=default_enddate.month)
    if zero_time:
        if not max_starttime:
            startdate = startdate.replace(
                hour=0, minute=0, second=0, microsecond=0
            )
        if not max_endtime:
            enddate = enddate.replace(
                hour=0, minute=0, second=0, microsecond=0
            )
    if max_starttime:
        startdate = startdate.replace(
            hour=23, minute=59, second=59, microsecond=999999
        )
    if max_endtime:
        enddate = enddate.replace(
            hour=23, minute=59, second=59, microsecond=999999
        )
    if timezone_handling == 0:
        startdate = startdate.replace(tzinfo=timezone.utc)
        enddate = enddate.replace(tzinfo=timezone.utc)
    elif timezone_handling == 3:
        if startdate.tzinfo is None:
            startdate = startdate.replace(tzinfo=timezone.utc)
        if enddate.tzinfo is None:
            enddate = enddate.replace(tzinfo=timezone.utc)
    elif timezone_handling == 4:
        startdate = startdate.astimezone(timezone.utc)
        enddate = enddate.astimezone(timezone.utc)
    elif timezone_handling == 5:
        if startdate.tzinfo is None:
            startdate = startdate.replace(tzinfo=timezone.utc)
        else:
            startdate = startdate.astimezone(timezone.utc)
        if enddate.tzinfo is None:
            enddate = enddate.replace(tzinfo=timezone.utc)
        else:
            enddate = enddate.astimezone(timezone.utc)
    if not include_microseconds:
        startdate = startdate.replace(microsecond=0)
        enddate = enddate.replace(microsecond=0)
    if fuzzy is not None:
        fuzzy["startdate"] = startdate
        fuzzy["enddate"] = enddate
    return startdate, enddate


def parse_date(
    string: str,
    date_format: Optional[str] = None,
    timezone_handling: int = 0,
    fuzzy: Optional[Dict] = None,
    include_microseconds: bool = False,
    zero_time: bool = False,
    max_time: bool = False,
    default_timezones: Optional[str] = None,
) -> datetime:
    """Parse date from string using specified date_format and return a datetime
    object. Raises exception for dates that are missing year, month or day. If
    no date_format is supplied, the function will guess, which for unambiguous
    formats, should work fine.

    By default, no timezone information will be parsed and the returned datetime will
    have timezone UTC. To change this behaviour, timezone_handling should be changed
    from its default of 0. If it is 1, then no timezone information will be parsed and a
    naive datetime will be returned. If it is 2 or more, then timezone information will
    be parsed. For 2, failure to parse timezone will result in a naive datetime. For 3,
    failure to parse timezone will result in the timezone being set to UTC. For 4 and 5,
    the time will be converted from whatever timezone is identified to UTC. For 4,
    failure to parse timezone will result in a naive (local) datetime converted to UTC.
    For 5, failure to parse timezone will result in the timezone being set to UTC.

    To parse a date within a string containing other text, you can supply a dictionary
    in the fuzzy parameter. In this case, dateutil's fuzzy parsing is used and the
    results returned in the dictionary in keys startdate, enddate, date (the string
    elements used to make the date) and nondate (the non date part of the string).

    By default, microseconds are ignored (set to 0), but can be included by setting
    include_microseconds to True. Any time elements are set to 0 if zero_time is True.
    If max_starttime is True, then the start date's time is set to 23:59:59. If
    max_endtime is True, then the end date's time is set to 23:59:59.

    When inferring time zones, a default set of time zones will be used unless
    overridden by passing in default_timezones which is a string of the form:

        -11 X NUT SST
        -10 W CKT HAST HST TAHT TKT

    Args:
        string (str): Dataset date string
        date_format (Optional[str]): Date format. If None is given, will attempt to guess. Defaults to None.
        timezone_handling (int): Timezone handling. See description. Defaults to 0 (ignore timezone, return UTC).
        fuzzy (Optional[Dict]): If dict supplied, fuzzy matching will be used and results returned in dict
        include_microseconds (bool): Includes microseconds if True. Defaults to False.
        zero_time (bool): Zero time elements of datetime if True. Defaults to False.
        max_time (bool): Make date time component 23:59:59:999999. Defaults to False.
        default_timezones (Optional[str]): Timezone information. Defaults to None. (Internal default).

    Returns:
        datetime: The parsed date
    """
    if max_time:
        max_starttime = True
        max_endtime = True
        zero_time = False
    else:
        max_starttime = False
        max_endtime = False

    startdate, enddate = parse_date_range(
        string,
        date_format=date_format,
        timezone_handling=timezone_handling,
        fuzzy=fuzzy,
        include_microseconds=include_microseconds,
        zero_time=zero_time,
        max_starttime=max_starttime,
        max_endtime=max_endtime,
        default_timezones=default_timezones,
    )
    if startdate != enddate:
        raise ParserError("date is not a specific day!")
    return startdate


def get_timestamp_from_datetime(date: datetime) -> float:
    """Convert datetime to timestamp.

    Args:
        date (datetime): Date to convert

    Returns:
        float: Timestamp
    """
    if date.tzinfo is None:
        return (
            time.mktime(
                (
                    date.year,
                    date.month,
                    date.day,
                    date.hour,
                    date.minute,
                    date.second,
                    -1,
                    -1,
                    -1,
                )
            )
            + date.microsecond / 1e6
        )
    else:
        return (
            date - datetime(1970, 1, 1, tzinfo=timezone.utc)
        ).total_seconds()


def get_datetime_from_timestamp(
    timestamp: float,
    timezone: datetime.tzinfo = timezone.utc,
    today: datetime = now_utc(),
) -> datetime:
    """Convert timestamp to datetime.

    Args:
        timestamp (float): Timestamp to convert
        timezone (datetime.tzinfo): Timezone to use
        today (datetime): Today's date. Defaults to now_utc.

    Returns:
        datetime: Date of timestamp
    """
    if timestamp > get_timestamp_from_datetime(today):
        timestamp = timestamp / 1000
    return datetime.fromtimestamp(timestamp, tz=timezone)


def iso_string_from_datetime(date: datetime) -> str:
    """Convert datetime to ISO formatted date without any time elements

    Args:
        date (datetime): Date to convert to string

    Returns:
        str: ISO formatted date without any time elements
    """
    return date.date().isoformat()



"""Saving utilities for YAML, JSON etc."""
import csv
import json
from collections import OrderedDict
from os.path import join
from typing import Any, Dict

from ruamel.yaml import (
    YAML,
    RoundTripRepresenter,
    SafeRepresenter,
    add_representer,
)

from .text import match_template_variables
from .typehint import ListTuple, ListTupleDict


class UnPrettyRTRepresenter(RoundTripRepresenter):
    def represent_none(self, data: Any) -> Any:
        return self.represent_scalar("tag:yaml.org,2002:null", "null")


class UnPrettySafeRepresenter(SafeRepresenter):
    def represent_none(self, data: Any) -> Any:
        return self.represent_scalar("tag:yaml.org,2002:null", "null")


class PrettySafeRepresenter(SafeRepresenter):
    def represent_none(self, data: Any) -> Any:
        if (
            len(self.represented_objects) == 0
            and not self.serializer.use_explicit_start
        ):
            # this will be open ended (although it is not yet)
            return self.represent_scalar("tag:yaml.org,2002:null", "null")
        return self.represent_scalar("tag:yaml.org,2002:null", "")


UnPrettyRTRepresenter.add_representer(
    None, UnPrettyRTRepresenter.represent_none
)
UnPrettySafeRepresenter.add_representer(
    None, UnPrettySafeRepresenter.represent_none
)
PrettySafeRepresenter.add_representer(
    None, PrettySafeRepresenter.represent_none
)


representers = {
    False: {False: UnPrettyRTRepresenter, True: RoundTripRepresenter},
    True: {False: UnPrettySafeRepresenter, True: PrettySafeRepresenter},
}


def save_text(string: str, path: str, encoding: str = "utf-8") -> None:
    """Save text string to file.

    Args:
        string (str): String to save
        path (str): Path to file
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        None
    """
    with open(path, "w", encoding=encoding) as f:
        f.write(string)


def save_yaml(
    object: Any,
    path: str,
    encoding: str = "utf-8",
    pretty: bool = False,
    sortkeys: bool = False,
) -> None:
    """Save dictionary to YAML file preserving order if it is an OrderedDict.

    Args:
        object (Any): Python object to save
        path (str): Path to YAML file
        encoding (str): Encoding of file. Defaults to utf-8.
        pretty (bool): Whether to pretty print. Defaults to False.
        sortkeys (bool): Whether to sort dictionary keys. Defaults to False.

    Returns:
        None
    """
    with open(path, "w", encoding=encoding) as f:
        representer = representers[sortkeys][pretty]
        yaml = YAML(typ="rt")
        yaml.Representer = representer
        add_representer(
            OrderedDict, representer.represent_dict, representer=representer
        )
        if pretty:
            yaml.indent(offset=2)
        else:
            yaml.default_flow_style = None
        yaml.representer.add_representer(
            type(None), representer.represent_none
        )
        yaml.dump(object, f)


def save_json(
    object: Any,
    path: str,
    encoding: str = "utf-8",
    pretty: bool = False,
    sortkeys: bool = False,
) -> None:
    """Save dictionary to JSON file preserving order if it is an OrderedDict.

    Args:
        object (Any): Python object to save
        path (str): Path to JSON file
        encoding (str): Encoding of file. Defaults to utf-8.
        pretty (bool): Whether to pretty print. Defaults to False.
        sortkeys (bool): Whether to sort dictionary keys. Defaults to False.

    Returns:
        None
    """
    with open(path, "w", encoding=encoding) as f:
        if pretty:
            indent = 2
            separators = (",", ": ")
        else:
            indent = None
            separators = (", ", ": ")
        json.dump(
            object,
            f,
            indent=indent,
            sort_keys=sortkeys,
            separators=separators,
        )


def save_hxlated_output(
    configuration: Dict,
    rows: ListTuple[ListTupleDict],
    includes_header: bool = True,
    includes_hxltags: bool = False,
    output_dir: str = "",
    **kwargs: Any,
) -> None:
    """Save rows with header and HXL hashtags. Currently, JSON and/or csv
    outputs are supported. An output directory (which defaults to "") can be
    given. The utility expects a configuration containing definitions of input
    headers and HXLtags if these are not included in the rows, and output files
    with desired HXL hashtags. Keyword arguments are used to pass in any
    variables needed by the metadata defined in the configuration.

    Args:
        configuration (Dict): Configuration for input and output
        rows (ListTuple[ListTupleDict]): Rows of data
        includes_header (bool): Whether rows includes header. Defaults to True,
        includes_hxltags (bool): Whether rows includes HXL hashtags. Defaults to False.
        output_dir (str): Output directory. Defaults to "".
        **kwargs: Variables to use when evaluating template arguments

    Returns:
        None
    """
    row0 = rows[0]
    if includes_header:
        if isinstance(row0, dict):
            headers = list(row0.keys())
        else:
            headers = rows[0]
            rows = rows[1:]
    else:
        headers = configuration["input"]["headers"]
    if includes_hxltags:
        if isinstance(row0, dict):
            hxltags = list(row0.values())
            rows = rows[1:]
        else:
            hxltags = rows[0]
            rows = rows[1:]
    else:
        hxltags = configuration["input"]["hxltags"]
    expressions = {}
    for process in configuration.get("process", []):
        headers.append(process["header"])
        hxltag = process["hxltag"]
        hxltags.append(hxltag)
        expressions[hxltag] = process["expression"]

    hxltag_to_header = dict(zip(hxltags, headers))
    csv_configuration = configuration["output"].get("csv")
    if csv_configuration:
        csv_hxltags = csv_configuration.get("hxltags", hxltags)
        csv_headers = [hxltag_to_header[hxltag] for hxltag in csv_hxltags]
        csv_file = open(
            join(output_dir, csv_configuration["filename"]),
            "w",
            encoding="utf-8",
            newline="\n",
        )
        output_csv = csv.writer(csv_file)
        output_csv.writerow(csv_headers)
        output_csv.writerow(csv_hxltags)
    else:
        csv_file = None
        output_csv = None
    json_configuration = configuration["output"].get("json")
    if json_configuration:
        data_key = json_configuration.get("data")
        json_hxltags = json_configuration.get("hxltags", hxltags)
        metadata_configuration = json_configuration.get("metadata")
        if metadata_configuration:
            metadata = {}
            for metadata_name, value in metadata_configuration.items():
                if isinstance(value, str):
                    template_string, match_string = match_template_variables(
                        value
                    )
                    if template_string:
                        value = kwargs.get(match_string)
                if value is None:
                    continue
                metadata[metadata_name] = value
            metadata_json = json.dumps(
                metadata, indent=None, separators=(",", ":")
            )
        else:
            metadata_json = None

        output_json = open(
            join(output_dir, json_configuration["filename"]), "w"
        )

        if metadata_json:
            metadata_key = metadata_configuration.get("key", "metadata")
            if data_key is None:
                data_key = "data"
            output_string = (
                f'{{"{metadata_key}":{metadata_json},"{data_key}":[\n'
            )
        elif data_key is None:
            output_string = "[\n"
        else:
            output_string = f'{{"{data_key}":[\n'
        output_json.write(output_string)
    else:
        output_json = None
        data_key = None

    def write_row(inrow, ending):
        if isinstance(inrow, dict):
            inrow = list(inrow.values())

        def get_outrow(file_hxltags):
            outrow = {}
            for file_hxltag in file_hxltags:
                expression = expressions.get(file_hxltag)
                if expression:
                    for i, hxltag in enumerate(hxltags):
                        expression = expression.replace(hxltag, f"inrow[{i}]")
                    outrow[file_hxltag] = eval(expression)
                else:
                    outrow[file_hxltag] = inrow[hxltags.index(file_hxltag)]
            return outrow

        if output_csv:
            output_csv.writerow(get_outrow(csv_hxltags).values())
        if output_json:
            row = get_outrow(json_hxltags)
            output_json.write(
                json.dumps(row, indent=None, separators=(",", ":")) + ending
            )

    [write_row(row, ",\n") for row in rows[:-1]]
    write_row(rows[-1], "\n]")
    if output_json:
        if data_key is not None:
            output_json.write("}")
        output_json.close()
    if csv_file:
        csv_file.close()



"""User agent utilities."""
import logging
import os
from os.path import expanduser, isfile, join
from typing import Any, Dict, Optional

from . import __version__
from .loader import load_yaml

logger = logging.getLogger(__name__)


class UserAgentError(Exception):
    pass


class UserAgent:
    default_user_agent_config_yaml = join(expanduser("~"), ".useragent.yml")
    user_agent = None

    @staticmethod
    def _environment_variables(**kwargs: Any) -> Any:
        """Overwrite keyword arguments with environment variables.

        Args:
            **kwargs: See below
            user_agent (str): User agent string.

        Returns:
            kwargs: Changed keyword arguments
        """
        user_agent = os.getenv("USER_AGENT")
        if user_agent is not None:
            kwargs["user_agent"] = user_agent
        preprefix = os.getenv("PREPREFIX")
        if preprefix is not None:
            kwargs["preprefix"] = preprefix
        return kwargs

    @staticmethod
    def _construct(configdict: Dict, prefix: str, ua: str) -> str:
        """Construct user agent.

        Args:
            configdict (str): Additional configuration for user agent
            prefix (str): Text to put at start of user agent
            ua (str): Custom user agent text

        Returns:
            str: Full user agent string
        """
        if not ua:
            raise UserAgentError(
                "User_agent parameter missing. It can be your project's name for example."
            )
        preprefix = configdict.get("preprefix")
        if preprefix:
            user_agent = f"{preprefix}:"
        else:
            user_agent = ""
        if prefix:
            user_agent = f"{user_agent}{prefix}-"
        user_agent = f"{user_agent}{ua}"
        return user_agent

    @classmethod
    def _load(
        cls,
        prefix: str,
        user_agent_config_yaml: str,
        user_agent_lookup: Optional[str] = None,
    ) -> str:
        """Load user agent YAML file.

        Args:
            prefix (str): Text to put at start of user agent
            user_agent_config_yaml (str): Path to user agent YAML file
            user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.

        Returns:
            str: user agent
        """
        if not user_agent_config_yaml:
            user_agent_config_yaml = cls.default_user_agent_config_yaml
            logger.info(
                f"No user agent or user agent config file given. Using default user agent config file: {user_agent_config_yaml}."
            )
        if not isfile(user_agent_config_yaml):
            raise UserAgentError(
                "User_agent should be supplied in a YAML config file. It can be your project's name for example."
            )
        logger.info(
            f"Loading user agent config from: {user_agent_config_yaml}"
        )
        user_agent_config_dict = load_yaml(user_agent_config_yaml)
        if user_agent_lookup:
            user_agent_config_dict = user_agent_config_dict.get(
                user_agent_lookup
            )
        if not user_agent_config_dict:
            raise UserAgentError(
                f"No user agent information read from: {user_agent_config_yaml}"
            )
        ua = user_agent_config_dict.get("user_agent")
        return cls._construct(user_agent_config_dict, prefix, ua)

    @classmethod
    def _create(
        cls,
        user_agent: Optional[str] = None,
        user_agent_config_yaml: Optional[str] = None,
        user_agent_lookup: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        """Get full user agent string.

        Args:
            user_agent (Optional[str]): User agent string. HDXPythonLibrary/X.X.X- is prefixed.
            user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.

        Returns:
            str: Full user agent string
        """
        kwargs = UserAgent._environment_variables(**kwargs)
        user_agent = kwargs.pop("user_agent", user_agent)
        prefix = kwargs.pop("prefix", None)
        if not prefix:
            prefix = f"HDXPythonUtilities/{__version__}"
        if not user_agent:
            ua = cls._load(prefix, user_agent_config_yaml, user_agent_lookup)
        else:
            ua = cls._construct(kwargs, prefix, user_agent)
        return ua

    @classmethod
    def clear_global(cls) -> None:
        """Clear stored user agent string.

        Returns:
            None
        """
        cls.user_agent = None

    @classmethod
    def set_global(
        cls,
        user_agent: Optional[str] = None,
        user_agent_config_yaml: Optional[str] = None,
        user_agent_lookup: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Set global user agent string.

        Args:
            user_agent (Optional[str]): User agent string. HDXPythonLibrary/X.X.X- is prefixed.
            user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.

        Returns:
            None
        """
        cls.user_agent = cls._create(
            user_agent, user_agent_config_yaml, user_agent_lookup, **kwargs
        )

    @classmethod
    def get(
        cls,
        user_agent: Optional[str] = None,
        user_agent_config_yaml: Optional[str] = None,
        user_agent_lookup: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        """Get full user agent string from parameters if supplied falling back
        on global user agent if set.

        Args:
            user_agent (Optional[str]): User agent string. HDXPythonLibrary/X.X.X- is prefixed.
            user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.

        Returns:
            str: Full user agent string
        """
        if (
            user_agent
            or user_agent_config_yaml
            or "user_agent" in UserAgent._environment_variables(**kwargs)
        ):
            return UserAgent._create(
                user_agent, user_agent_config_yaml, user_agent_lookup, **kwargs
            )
        if cls.user_agent:
            return cls.user_agent
        else:
            raise UserAgentError(
                "You must either set the global user agent: UserAgent.set_global(...) or pass in user agent parameters!"
            )



"""Downloader Tests"""
import copy
import re
from collections import OrderedDict
from contextlib import contextmanager
from os import remove
from os.path import abspath, join
from shutil import copytree, rmtree
from tempfile import gettempdir

import pytest
from utils import assert_downloaders

from hdx.utilities.base_downloader import DownloadError
from hdx.utilities.downloader import Download
from hdx.utilities.session import SessionError
from hdx.utilities.useragent import UserAgent


@contextmanager
def not_raises(ExpectedException):
    try:
        yield
    except ExpectedException as error:
        raise pytest.fail(f"Raised exception {error} when it should not!")

    except Exception as error:
        raise pytest.fail(f"An unexpected exception {error} raised.")


class TestDownloader:
    downloaderfoldername = "downloader"

    @pytest.fixture(scope="class", autouse=True)
    def useragent(self):
        UserAgent.set_global("test")
        yield
        UserAgent.clear_global()

    @pytest.fixture(scope="class")
    def downloaderfolder(self, fixturesfolder):
        return join(fixturesfolder, self.downloaderfoldername)

    @pytest.fixture(scope="class")
    def fixturefile(self, downloaderfolder):
        return join(downloaderfolder, "extra_params_tree.yml")

    @pytest.fixture(scope="class")
    def fixtureurl(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/test_data.csv"

    @pytest.fixture(scope="class")
    def fixtureurlexcel(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/downloader/test_data.xlsx"

    @pytest.fixture(scope="class")
    def getfixtureurl(self):
        return "http://httpbin.org/get"

    @pytest.fixture(scope="class")
    def postfixtureurl(self):
        return "http://httpbin.org/post"

    @pytest.fixture(scope="class")
    def fixturenotexistsurl(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/NOTEXIST.csv"

    @pytest.fixture(scope="class")
    def fixtureprocessurl(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/downloader/test_csv_processing.csv?a=1"

    @pytest.fixture(scope="class")
    def fixtureprocessurlblank(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/downloader/test_csv_processing_blanks.csv?a=1"

    @pytest.fixture(scope="class")
    def fixturexlsurl(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/downloader/test_xls_processing.xls?a=1"

    @pytest.fixture(scope="class")
    def fixturexlsxurl(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/downloader/test_xlsx_processing.xlsx?a=1"

    @pytest.fixture(scope="class")
    def fixturejsonurl(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/downloader/test_json_processing.json?a=1"

    def test_get_path_for_url(
        self, tmpdir, fixtureurl, configfolder, downloaderfolder
    ):
        tmpdir = str(tmpdir)
        filename = "test_data.csv"
        path = Download.get_path_for_url(fixtureurl, configfolder)
        assert abspath(path) == abspath(join(configfolder, filename))
        path = Download.get_path_for_url(fixtureurl, downloaderfolder)
        assert abspath(path) == abspath(
            join(downloaderfolder, "test_data3.csv")
        )
        testfolder = join(tmpdir, self.downloaderfoldername)
        rmtree(testfolder, ignore_errors=True)
        copytree(downloaderfolder, testfolder)
        path = Download.get_path_for_url(
            fixtureurl, testfolder, overwrite=True
        )
        assert abspath(path) == abspath(join(testfolder, filename))
        rmtree(testfolder)
        filename = "myfilename.txt"
        path = Download.get_path_for_url(fixtureurl, filename=filename)
        assert abspath(path) == abspath(join(gettempdir(), filename))
        path = Download.get_path_for_url(
            fixtureurl, downloaderfolder, filename
        )
        assert abspath(path) == abspath(join(downloaderfolder, filename))

    def test_init(self, monkeypatch, downloaderfolder):
        basicauthfile = join(downloaderfolder, "basicauth.txt")
        with Download(auth=("u", "p")) as downloader:
            assert downloader.session.auth == ("u", "p")
        basicauth = "Basic dXNlcjpwYXNz"
        with Download(basic_auth=basicauth) as downloader:
            assert downloader.session.auth == ("user", "pass")
        with Download(basic_auth_file=basicauthfile) as downloader:
            assert downloader.session.auth == ("testuser", "testpass")
        extraparamsyamltree = join(downloaderfolder, "extra_params_tree.yml")
        with Download(
            extra_params_yaml=extraparamsyamltree, extra_params_lookup="mykey"
        ) as downloader:
            assert downloader.session.auth == ("testuser", "testpass")
        monkeypatch.setenv("BASIC_AUTH", basicauth)
        with Download() as downloader:
            assert downloader.session.auth == ("user", "pass")
        with pytest.raises(SessionError):
            Download(basic_auth="12345")
        with pytest.raises(SessionError):
            Download(
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="mykey",
            )
        monkeypatch.delenv("BASIC_AUTH")
        with pytest.raises(SessionError):
            Download(
                basic_auth=basicauth,
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="mykey",
            )
        with pytest.raises(SessionError):
            Download(auth=("u", "p"), basic_auth="Basic xxxxxxxxxxxxxxxx")
        extraparamsjson = join(downloaderfolder, "extra_params.json")
        with pytest.raises(SessionError):
            Download(auth=("u", "p"), basic_auth_file=extraparamsjson)
        with pytest.raises(SessionError):
            Download(
                basic_auth="Basic dXNlcjpwYXNz",
                basic_auth_file=extraparamsjson,
            )
        with pytest.raises(SessionError):
            Download(
                auth=("u", "p"),
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="mykey",
            )
        with pytest.raises(SessionError):
            Download(
                basic_auth_file=basicauthfile,
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="mykey",
            )
        with pytest.raises(SessionError):
            Download(
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="missingkey",
            )
        with pytest.raises(IOError):
            Download(basic_auth_file="NOTEXIST")
        extraparamsyaml = join(downloaderfolder, "extra_params.yml")
        test_url = "http://www.lalala.com/lala"
        with Download(
            basic_auth_file=basicauthfile, extra_params_dict={"key1": "val1"}
        ) as downloader:
            assert downloader.session.auth == ("testuser", "testpass")
            assert downloader.get_full_url(test_url) == f"{test_url}?key1=val1"
        key = "Authorization"
        value = "lala"
        with Download(headers={key: value}) as downloader:
            assert key in downloader.session.headers
            assert downloader.session.headers.get(key) == value
        with Download(extra_params_json=extraparamsjson) as downloader:
            full_url = downloader.get_full_url(test_url)
            assert "param_1=value+1" in full_url
            assert "param_2=value_2" in full_url
            assert "param_3=12" in full_url
        with Download(extra_params_yaml=extraparamsyaml) as downloader:
            full_url = downloader.get_full_url(test_url)
            assert "param1=value1" in full_url
            assert "param2=value+2" in full_url
            assert "param3=10" in full_url
        with Download(
            extra_params_yaml=extraparamsyamltree, extra_params_lookup="mykey"
        ) as downloader:
            assert downloader.session.auth == ("testuser", "testpass")
            full_url = downloader.get_full_url(test_url)
            assert "param1=value+1" in full_url
            assert "param2=value2" in full_url
            assert "param3=11" in full_url
            assert "basic_auth" not in full_url
        monkeypatch.setenv("EXTRA_PARAMS", "param1=value2+3,param2=value3+7")
        with Download(
            extra_params_yaml=extraparamsyamltree, extra_params_lookup="mykey"
        ) as downloader:
            assert downloader.session.auth is None
            full_url = downloader.get_full_url(test_url)
            assert "param1=value2%2B3" in full_url
            assert "param2=value3%2B7" in full_url
            assert "param3=11" not in full_url
            assert "basic_auth" not in full_url
        with Download(
            extra_params_yaml=extraparamsyamltree,
            extra_params_lookup="mykey",
            use_env=False,
        ) as downloader:
            assert downloader.session.auth == ("testuser", "testpass")
            full_url = downloader.get_full_url(test_url)
            assert "param1=value+1" in full_url
            assert "param2=value2" in full_url
            assert "param3=11" in full_url
            assert "basic_auth" not in full_url
        monkeypatch.delenv("EXTRA_PARAMS")
        with pytest.raises(SessionError):
            Download(
                extra_params_dict={"key1": "val1"},
                extra_params_json=extraparamsjson,
            )
        with pytest.raises(SessionError):
            Download(
                extra_params_dict={"key1": "val1"},
                extra_params_yaml=extraparamsyaml,
            )
        with pytest.raises(SessionError):
            Download(
                extra_params_dict={"key1": "val1"},
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="mykey",
            )
        with pytest.raises(IOError):
            Download(extra_params_json="NOTEXIST")
        with pytest.raises(IOError):
            Download(extra_params_yaml="NOTEXIST")
        with not_raises(IOError):
            Download(extra_params_json="NOTEXIST", fail_on_missing_file=False)
            Download(extra_params_yaml="NOTEXIST", fail_on_missing_file=False)
            Download(basic_auth_file="NOTEXIST", fail_on_missing_file=False)

    def test_get_url_for_get(self):
        assert (
            Download.get_url_for_get(
                "http://www.lala.com/hdfa?a=3&b=4",
                OrderedDict([("c", "e"), ("d", "f")]),
            )
            == "http://www.lala.com/hdfa?a=3&b=4&c=e&d=f"
        )
        assert (
            Download.get_url_for_get(
                "http://www.lala.com/hdfa?a=3&b=4", {"c": "e", "d": "f"}
            )
            == "http://www.lala.com/hdfa?a=3&b=4&c=e&d=f"
        )

    def test_get_url_params_for_post(self):
        result = Download.get_url_params_for_post(
            "http://www.lala.com/hdfa?a=3&b=4",
            OrderedDict([("c", "e"), ("d", "f")]),
        )
        assert result[0] == "http://www.lala.com/hdfa"
        assert list(result[1].items()) == list(
            OrderedDict(
                [("a", "3"), ("b", "4"), ("c", "e"), ("d", "f")]
            ).items()
        )
        result = Download.get_url_params_for_post(
            "http://www.lala.com/hdfa?a=3&b=4", {"c": "e", "d": "f"}
        )
        assert result[0] == "http://www.lala.com/hdfa"
        assert list(result[1].items()) == list(
            OrderedDict(
                [("a", "3"), ("b", "4"), ("c", "e"), ("d", "f")]
            ).items()
        )

    def test_hxl_row(self):
        headers = ["a", "b", "c"]
        hxltags = {"b": "#b", "c": "#c"}
        assert Download.hxl_row(headers, hxltags) == ["", "#b", "#c"]
        assert Download.hxl_row(headers, hxltags, dict_form=True) == {
            "a": "",
            "b": "#b",
            "c": "#c",
        }
        assert Download.hxl_row(headers, dict()) == ["", "", ""]
        assert Download.hxl_row([], hxltags) == list()

    def test_setup_stream(
        self, fixtureurl, fixturenotexistsurl, getfixtureurl, postfixtureurl
    ):
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.setup("NOTEXIST://NOTEXIST.csv")
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.setup(fixturenotexistsurl)
        with Download() as downloader:
            downloader.setup(fixtureurl)
            headers = downloader.response.headers
            assert bool(re.match(r"7\d\d", headers["Content-Length"])) is True
        with Download() as downloader:
            downloader.setup(postfixtureurl, post=True)
            headers = downloader.response.headers
            assert bool(re.match(r"4\d\d", headers["Content-Length"])) is True
            downloader.setup(
                f"{getfixtureurl}?id=10&lala=a",
                post=False,
                parameters=OrderedDict([("b", "4"), ("d", "3")]),
            )
            assert list(downloader.get_json()["args"].items()) == list(
                OrderedDict(
                    [("b", "4"), ("d", "3"), ("id", "10"), ("lala", "a")]
                ).items()
            )
            downloader.setup(
                f"{postfixtureurl}?id=3&lala=b",
                post=True,
                parameters=OrderedDict([("a", "3"), ("c", "2")]),
            )
            assert list(downloader.get_json()["form"].items()) == list(
                OrderedDict(
                    [("a", "3"), ("c", "2"), ("id", "3"), ("lala", "b")]
                ).items()
            )

    def test_hash_stream(self, fixtureurl):
        with Download() as downloader:
            downloader.setup(fixtureurl)
            md5hash = downloader.hash_stream(fixtureurl)
            assert md5hash == "da9db35a396cca10c618f6795bdb9ff2"

    def test_download_file(
        self,
        tmpdir,
        fixturefile,
        fixtureurl,
        fixturenotexistsurl,
        getfixtureurl,
        postfixtureurl,
    ):
        tmpdir = str(tmpdir)
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download_file("NOTEXIST://NOTEXIST.csv", folder=tmpdir)
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download_file(fixturenotexistsurl)
        filename = "myfilename.txt"
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download_file(
                fixturefile, folder=tmpdir, path=join(tmpdir, filename)
            )
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download_file(
                fixturefile, filename=filename, path=join(tmpdir, filename)
            )
        with Download() as downloader:
            f = downloader.download_file(fixturefile, folder=tmpdir)
            fpath = abspath(f)
            remove(f)
            assert fpath == abspath(join(tmpdir, "extra_params_tree.yml"))
            f = downloader.download_file(fixtureurl, folder=tmpdir)
            fpath = abspath(f)
            remove(f)
            assert fpath == abspath(join(tmpdir, "test_data.csv"))
            f = downloader.download_file(
                fixtureurl, folder=tmpdir, filename=filename
            )
            fpath = abspath(f)
            assert fpath == abspath(join(tmpdir, filename))
            f = downloader.download_file(
                fixtureurl, path=join(tmpdir, filename), overwrite=True
            )
            fpath = abspath(f)
            assert fpath == abspath(join(tmpdir, filename))
            f = downloader.download_file(
                fixtureurl, path=join(tmpdir, filename), overwrite=False
            )
            fpath = abspath(f)
            assert fpath == abspath(
                join(tmpdir, filename.replace(".txt", "1.txt"))
            )
            f = downloader.download_file(
                fixtureurl, path=join(tmpdir, filename), keep=True
            )
            fpath = abspath(f)
            remove(f)
            assert fpath == abspath(join(tmpdir, filename))
            f = downloader.download_file(
                f"{getfixtureurl}?id=10&lala=a",
                post=False,
                parameters=OrderedDict([("b", "4"), ("d", "3")]),
                folder=tmpdir,
                filename=filename,
            )
            fpath = abspath(f)
            with open(fpath) as fi:
                text = fi.read()
                assert '"id": "10"' in text
                assert '"lala": "a"' in text
                assert '"b": "4"' in text
                assert '"d": "3"' in text
            remove(f)
            assert fpath == abspath(join(tmpdir, filename))
            f = downloader.download_file(
                f"{postfixtureurl}?id=3&lala=b",
                post=True,
                parameters=OrderedDict([("a", "3"), ("c", "2")]),
                folder=tmpdir,
                filename=filename,
            )
            fpath = abspath(f)
            with open(fpath) as fi:
                text = fi.read()
                assert '"id": "3"' in text
                assert '"lala": "b"' in text
                assert '"a": "3"' in text
                assert '"c": "2"' in text
            remove(f)
            assert fpath == abspath(join(tmpdir, filename))

    def test_download(
        self,
        fixturefile,
        fixtureurl,
        fixturenotexistsurl,
        getfixtureurl,
        postfixtureurl,
    ):
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download("NOTEXIST://NOTEXIST.csv")
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download(fixturenotexistsurl)
        with Download() as downloader:
            text = downloader.download_text(fixtureurl)
            assert downloader.get_status() == 200
            assert len(downloader.get_headers()) in (24, 25)
            assert (
                bool(
                    re.match(r"7\d\d", downloader.get_header("Content-Length"))
                )
                is True
            )
            assert (
                text
                == """GWNO,EVENT_ID_CNTY,EVENT_ID_NO_CNTY,EVENT_DATE,YEAR,TIME_PRECISION,EVENT_TYPE,ACTOR1,ALLY_ACTOR_1,INTER1,ACTOR2,ALLY_ACTOR_2,INTER2,INTERACTION,COUNTRY,ADMIN1,ADMIN2,ADMIN3,LOCATION,LATITUDE,LONGITUDE,GEO_PRECISION,SOURCE,NOTES,FATALITIES
615,1416RTA,,18/04/2001,2001,1,Violence against civilians,Police Forces of Algeria (1999-),,1,Civilians (Algeria),Berber Ethnic Group (Algeria),7,17,Algeria,Tizi Ouzou,Beni-Douala,,Beni Douala,36.61954,4.08282,1,Associated Press Online,A Berber student was shot while in police custody at a police station in Beni Douala. He later died on Apr.21.,1
615,2229RTA,,19/04/2001,2001,1,Riots/Protests,Rioters (Algeria),Berber Ethnic Group (Algeria),5,Police Forces of Algeria (1999-),,1,15,Algeria,Tizi Ouzou,Tizi Ouzou,,Tizi Ouzou,36.71183,4.04591,3,Kabylie report,"Riots were reported in numerous villages in Kabylie, resulting in dozens wounded in clashes between protesters and police and significant material damage.",0
615,2230RTA,,20/04/2001,2001,1,Riots/Protests,Protesters (Algeria),Students (Algeria),6,,,0,60,Algeria,Bejaia,Amizour,,Amizour,36.64022,4.90131,1,Crisis Group,Students protested in the Amizour area. At least 3 were later arrested for allegedly insulting gendarmes.,0
615,2231RTA,,21/04/2001,2001,1,Riots/Protests,Rioters (Algeria),Berber Ethnic Group (Algeria),5,Police Forces of Algeria (1999-),,1,15,Algeria,Bejaia,Amizour,,Amizour,36.64022,4.90131,1,Kabylie report,"Rioters threw molotov cocktails, rocks and burning tires at gendarmerie stations in Beni Douala, El-Kseur and Amizour.",0

"""
            )
            downloader.download(
                f"{getfixtureurl}?id=10&lala=a",
                post=False,
                parameters=OrderedDict([("b", "4"), ("d", "3")]),
            )
            assert list(downloader.get_json()["args"].items()) == list(
                OrderedDict(
                    [("b", "4"), ("d", "3"), ("id", "10"), ("lala", "a")]
                ).items()
            )
            json = downloader.download_json(
                f"{postfixtureurl}?id=3&lala=b",
                post=True,
                parameters=OrderedDict([("a", "3"), ("c", "2")]),
            )
            assert list(json["form"].items()) == list(
                OrderedDict(
                    [("a", "3"), ("c", "2"), ("id", "3"), ("lala", "b")]
                ).items()
            )
        with Download(rate_limit={"calls": 1, "period": 0.1}) as downloader:
            downloader.download(
                f"{getfixtureurl}?id=10&lala=a",
                post=False,
                parameters=OrderedDict([("b", "4"), ("d", "3")]),
            )
            assert list(downloader.get_json()["args"].items()) == list(
                OrderedDict(
                    [("b", "4"), ("d", "3"), ("id", "10"), ("lala", "a")]
                ).items()
            )
            yaml = downloader.download_yaml(fixturefile)
            assert list(yaml["mykey"].items()) == list(
                OrderedDict(
                    [
                        ("param1", "value 1"),
                        ("param2", "value2"),
                        ("param3", 11),
                        ("basic_auth", "Basic dGVzdHVzZXI6dGVzdHBhc3M="),
                    ]
                ).items()
            )

    def test_download_tabular_key_value(
        self, fixtureurl, fixtureurlexcel, fixtureprocessurl
    ):
        with Download() as downloader:
            result = downloader.download_tabular_key_value(
                fixtureurl,
                file_type="csv",
                delimiter=",",
                skip_initial_space=True,
                infer_types=True,
            )
            assert result == {615: "2231RTA", "GWNO": "EVENT_ID_CNTY"}
            result = downloader.download_tabular_key_value(
                fixtureurlexcel,
                file_type="xlsx",
                sheet="test_data",
                infer_types=True,
            )
            assert result == {615: "2231RTA", "GWNO": "EVENT_ID_CNTY"}
            result = downloader.download_tabular_key_value(
                fixtureprocessurl, headers=3
            )
            assert result == {"coal": "3", "gas": "2"}
            with pytest.raises(DownloadError):
                downloader.download_tabular_key_value(
                    "NOTEXIST://NOTEXIST.csv"
                )

    def test_get_tabular_rows_as_list(self, fixtureprocessurl):
        with Download() as downloader:
            headers, rows = downloader.get_tabular_rows_as_list(
                fixtureprocessurl
            )
            rows = list(rows)
            assert rows == [
                ["la1", "ha1", "ba1", "ma1"],
                ["header1", "header2", "header3", "header4"],
                ["coal", "3", "7.4", "'needed'"],
                ["gas", "2", "6.5", "'n/a'"],
            ]

    def test_get_tabular_rows_as_dict(self, fixtureprocessurl):
        with Download() as downloader:
            headers, rows = downloader.get_tabular_rows_as_dict(
                fixtureprocessurl
            )
            rows = list(rows)
            assert rows == [
                {
                    "la1": "header1",
                    "ha1": "header2",
                    "ba1": "header3",
                    "ma1": "header4",
                },
                {"la1": "coal", "ha1": "3", "ba1": "7.4", "ma1": "'needed'"},
                {"la1": "gas", "ha1": "2", "ba1": "6.5", "ma1": "'n/a'"},
            ]

    def test_get_tabular_rows(self, fixtureprocessurl, fixtureprocessurlblank):
        with Download() as downloader:
            expected = [
                ["la1", "ha1", "ba1", "ma1"],
                ["header1", "header2", "header3", "header4"],
                ["coal", "3", "7.4", "'needed'"],
                ["gas", "2", "6.5", "'n/a'"],
            ]
            expected_headers = expected[0]
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl)
            assert headers == expected_headers
            assert list(iterator) == expected[1:]
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank
            )
            assert headers == expected_headers
            blank_expected = copy.deepcopy(expected[1:])
            blank_expected[2][0] = None
            assert list(iterator) == blank_expected
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank, ignore_blank_rows=False
            )
            assert headers == expected_headers
            blank_expected.append([None, None, None, None])
            blank_expected.insert(1, [None, None, None, None])
            assert list(iterator) == blank_expected
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=1
            )
            assert headers == expected_headers
            assert list(iterator) == expected[1:]
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=2
            )
            assert headers == expected[1]
            assert list(iterator) == expected[2:]
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=[1, 2]
            )
            assert headers == [
                "la1 header1",
                "ha1 header2",
                "ba1 header3",
                "ma1 header4",
            ]
            assert list(iterator) == expected[2:]
            myheaders = ["a", "b", "c", "d"]
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=myheaders
            )
            assert headers == myheaders
            assert list(iterator) == expected
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=1, dict_form=True
            )
            assert headers == expected_headers
            expected_dicts = [
                {
                    "la1": "header1",
                    "ha1": "header2",
                    "ba1": "header3",
                    "ma1": "header4",
                },
                {"la1": "coal", "ha1": "3", "ba1": "7.4", "ma1": "'needed'"},
                {"la1": "gas", "ha1": "2", "ba1": "6.5", "ma1": "'n/a'"},
            ]
            assert list(iterator) == expected_dicts
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=3, dict_form=True
            )
            assert headers == expected[2]
            expected_dicts = [
                {"coal": "gas", "3": "2", "7.4": "6.5", "'needed'": "'n/a'"}
            ]
            assert list(iterator) == expected_dicts

            def testfn(headers, row):
                row.insert(2, "lala")
                return row

            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl,
                headers=3,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            expected_headers = ["coal", "3", "la", "7.4", "'needed'"]
            assert headers == expected_headers
            assert list(iterator) == [["gas", "2", "lala", "6.5", "'n/a'"]]

            def testfn(headers, row):
                row["la"] = "lala"
                return row

            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl,
                headers=3,
                dict_form=True,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            assert headers == expected_headers
            expected_dicts[0]["la"] = "lala"
            assert list(iterator) == expected_dicts

            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank,
                headers=1,
                dict_form=True,
                ignore_blank_rows=False,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            expected_headers = ["la1", "ha1", "la", "ba1", "ma1"]
            assert headers == expected_headers

            expected_dicts = [
                {
                    "la1": "header1",
                    "ha1": "header2",
                    "ba1": "header3",
                    "ma1": "header4",
                    "la": "lala",
                },
                {
                    "la1": None,
                    "ha1": None,
                    "ba1": None,
                    "ma1": None,
                    "la": "lala",
                },
                {
                    "la1": "coal",
                    "ha1": "3",
                    "ba1": "7.4",
                    "ma1": "'needed'",
                    "la": "lala",
                },
                {
                    "la1": None,
                    "ha1": "2",
                    "ba1": "6.5",
                    "ma1": "'n/a'",
                    "la": "lala",
                },
                {
                    "la1": None,
                    "ha1": None,
                    "ba1": None,
                    "ma1": None,
                    "la": "lala",
                },
            ]
            assert list(iterator) == expected_dicts
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank,
                headers=1,
                dict_form=True,
                ignore_blank_rows=False,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            assert headers == expected_headers
            assert list(iterator) == expected_dicts
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank,
                headers=1,
                dict_form=True,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            assert headers == expected_headers
            del expected_dicts[4]
            del expected_dicts[1]
            assert list(iterator) == expected_dicts

            def testfn(headers, row):
                row["la"] = "lala"
                if row["ha1"] == "3":
                    return None
                return row

            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank,
                headers=1,
                dict_form=True,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            assert headers == expected_headers
            del expected_dicts[1]
            assert list(iterator) == expected_dicts

            with pytest.raises(DownloadError):
                downloader.get_tabular_rows(fixtureprocessurl, headers=None)

    def test_get_tabular_rows_xls(self, fixturexlsurl):
        with Download() as downloader:
            headers, iterator = downloader.get_tabular_rows(
                fixturexlsurl, format="xls", headers=4
            )
            assert headers == [
                "Country Name",
                "Country Code",
                "Indicator Name",
                "Indicator Code",
                "Year",
                "Value",
            ]
            rows = list(iterator)
            assert rows[0] == [
                "Aruba",
                "ABW",
                "Population, total",
                "SP.POP.TOTL",
                1960,
                54211,
            ]
            assert rows[-1] == [
                "Zimbabwe",
                "ZWE",
                "Population, total",
                "SP.POP.TOTL",
                2019,
                14645468,
            ]

    def test_get_tabular_rows_xlsx(self, fixturexlsxurl):
        with Download() as downloader:
            headers, iterator = downloader.get_tabular_rows(
                fixturexlsxurl, format="xlsx", sheet="Trend", headers=[7, 8]
            )
            assert headers[1:222] == [
                "ISO",
                "Countries and areas",
                "UNICEF Regions",
                "UN Regions",
                "WHO Regions",
                "World Bank Income Groups",
                "Survey Years",
                "Year*",
                "Short Source",
                "Full Source Title",
                "Latest Estimate",
                "Estimate Type",
                "National Point Estimate",
                "National Lower Limit",
                "National Upper Limit",
                "National Sample Size",
                "National Footnote",
                "Region 1 Point Estimate",
                "Region 1 Lower Limit",
                "Region 1 Upper Limit",
                "Region 1 Sample Size",
                "Region 1 Region Name",
                "Region 1 Footnote",
                "Region 2 Point Estimate",
                "Region 2 Lower Limit",
                "Region 2 Upper Limit",
                "Region 2 Sample Size",
                "Region 2 Region Name",
                "Region 2 Footnote",
                "Region 3 Point Estimate",
                "Region 3 Lower Limit",
                "Region 3 Upper Limit",
                "Region 3 Sample Size",
                "Region 3 Region Name",
                "Region 3 Footnote",
                "Region 4 Point Estimate",
                "Region 4 Lower Limit",
                "Region 4 Upper Limit",
                "Region 4 Sample Size",
                "Region 4 Region Name",
                "Region 4 Footnote",
                "Region 5 Point Estimate",
                "Region 5 Lower Limit",
                "Region 5 Upper Limit",
                "Region 5 Sample Size",
                "Region 5 Region Name",
                "Region 5 Footnote",
                "Region 6 Point Estimate",
                "Region 6 Lower Limit",
                "Region 6 Upper Limit",
                "Region 6 Sample Size",
                "Region 6 Region Name",
                "Region 6 Footnote",
                "Region 7 Point Estimate",
                "Region 7 Lower Limit",
                "Region 7 Upper Limit",
                "Region 7 Sample Size",
                "Region 7 Region Name",
                "Region 7 Footnote",
                "Region 8 Point Estimate",
                "Region 8 Lower Limit",
                "Region 8 Upper Limit",
                "Region 8 Sample Size",
                "Region 8 Region Name",
                "Region 8 Footnote",
                "Region 9 Point Estimate",
                "Region 9 Lower Limit",
                "Region 9 Upper Limit",
                "Region 9 Sample Size",
                "Region 9 Region Name",
                "Region 9 Footnote",
                "Region 10 Point Estimate",
                "Region 10 Lower Limit",
                "Region 10 Upper Limit",
                "Region 10 Sample Size",
                "Region 10 Region Name",
                "Region 10 Footnote",
                "Region 11 Point Estimate",
                "Region 11 Lower Limit",
                "Region 11 Upper Limit",
                "Region 11 Sample Size",
                "Region 11 Region Name",
                "Region 11 Footnote",
                "Region 12 Point Estimate",
                "Region 12 Lower Limit",
                "Region 12 Upper Limit",
                "Region 12 Sample Size",
                "Region 12 Region Name",
                "Region 12 Footnote",
                "Region 13 Point Estimate",
                "Region 13 Lower Limit",
                "Region 13 Upper Limit",
                "Region 13 Sample Size",
                "Region 13 Region Name",
                "Region 13 Footnote",
                "Region 14 Point Estimate",
                "Region 14 Lower Limit",
                "Region 14 Upper Limit",
                "Region 14 Sample Size",
                "Region 14 Region Name",
                "Region 14 Footnote",
                "Region 15 Point Estimate",
                "Region 15 Lower Limit",
                "Region 15 Upper Limit",
                "Region 15 Sample Size",
                "Region 15 Region Name",
                "Region 15 Footnote",
                "Region 16 Point Estimate",
                "Region 16 Lower Limit",
                "Region 16 Upper Limit",
                "Region 16 Sample Size",
                "Region 16 Region Name",
                "Region 16 Footnote",
                "Region 17 Point Estimate",
                "Region 17 Lower Limit",
                "Region 17 Upper Limit",
                "Region 17 Sample Size",
                "Region 17 Region Name",
                "Region 17 Footnote",
                "Region 18 Point Estimate",
                "Region 18 Lower Limit",
                "Region 18 Upper Limit",
                "Region 18 Sample Size",
                "Region 18 Region Name",
                "Region 18 Footnote",
                "Region 19 Point Estimate",
                "Region 19 Lower Limit",
                "Region 19 Upper Limit",
                "Region 19 Sample Size",
                "Region 19 Region Name",
                "Region 19 Footnote",
                "Region 20 Point Estimate",
                "Region 20 Lower Limit",
                "Region 20 Upper Limit",
                "Region 20 Sample Size",
                "Region 20 Region Name",
                "Region 20 Footnote",
                "Region 21 Point Estimate",
                "Region 21 Lower Limit",
                "Region 21 Upper Limit",
                "Region 21 Sample Size",
                "Region 21 Region Name",
                "Region 21 Footnote",
                "Region 22 Point Estimate",
                "Region 22 Lower Limit",
                "Region 22 Upper Limit",
                "Region 22 Sample Size",
                "Region 22 Region Name",
                "Region 22 Footnote",
                "Region 23 Point Estimate",
                "Region 23 Lower Limit",
                "Region 23 Upper Limit",
                "Region 23 Sample Size",
                "Region 23 Region Name",
                "Region 23 Footnote",
                "Region 24 Point Estimate",
                "Region 24 Lower Limit",
                "Region 24 Upper Limit",
                "Region 24 Sample Size",
                "Region 24 Region Name",
                "Region 24 Footnote",
                "Region 25 Point Estimate",
                "Region 25 Lower Limit",
                "Region 25 Upper Limit",
                "Region 25 Sample Size",
                "Region 25 Region Name",
                "Region 25 Footnote",
                "Region 26 Point Estimate",
                "Region 26 Lower Limit",
                "Region 26 Upper Limit",
                "Region 26 Sample Size",
                "Region 26 Region Name",
                "Region 26 Footnote",
                "Region 27 Point Estimate",
                "Region 27 Lower Limit",
                "Region 27 Upper Limit",
                "Region 27 Sample Size",
                "Region 27 Region Name",
                "Region 27 Footnote",
                "Region 28 Point Estimate",
                "Region 28 Lower Limit",
                "Region 28 Upper Limit",
                "Region 28 Sample Size",
                "Region 28 Region Name",
                "Region 28 Footnote",
                "Region 29 Point Estimate",
                "Region 29 Lower Limit",
                "Region 29 Upper Limit",
                "Region 29 Sample Size",
                "Region 29 Region Name",
                "Region 29 Footnote",
                "Region 30 Point Estimate",
                "Region 30 Lower Limit",
                "Region 30 Upper Limit",
                "Region 30 Sample Size",
                "Region 30 Region Name",
                "Region 30 Footnote",
                "Region 31 Point Estimate",
                "Region 31 Lower Limit",
                "Region 31 Upper Limit",
                "Region 31 Sample Size",
                "Region 31 Region Name",
                "Region 31 Footnote",
                "Region 32 Point Estimate",
                "Region 32 Lower Limit",
                "Region 32 Upper Limit",
                "Region 32 Sample Size",
                "Region 32 Region Name",
                "Region 32 Footnote",
                "Region 33 Point Estimate",
                "Region 33 Lower Limit",
                "Region 33 Upper Limit",
                "Region 33 Sample Size",
                "Region 33 Region Name",
                "Region 33 Footnote",
                "Region 34 Point Estimate",
                "Region 34 Lower Limit",
                "Region 34 Upper Limit",
                "Region 34 Sample Size",
                "Region 34 Region Name",
                "Region 34 Footnote",
            ]
            rows = list(iterator)
            assert rows[0][1:17] == [
                "AFG",
                "Afghanistan",
                "SA",
                "Asia",
                "EMRO",
                "Low Income",
                "2004",
                2004,
                "NNS",
                "Summary report of the national nutrition survey, 2004. Kabul, Islamic "
                "Republic of Afghanistan: Ministry of Public Health and UNICEF, 2005.",
                "Historical Trend",
                "External Reanalysis",
                3.5,
                None,
                None,
                946,
            ]
            assert rows[-1][1:113] == [
                "PSE",
                "State of Palestine",
                "MENA",
                "Asia",
                "Not Classified",
                "Lower Middle Income",
                "2014",
                2014,
                "MICS",
                "Palestinian Multiple Indicator Cluster Survey 2014,  Final Report, Ramallah, "
                "Palestine",
                "Latest Source",
                "Reanalyzed",
                0.3240246,
                0.2182891,
                0.4807297,
                6906.391,
                None,
                0.9850393,
                0.3510135,
                2.73288,
                274.0501,
                "Bethlehem",
                None,
                0.2285949,
                0.0331941,
                1.556335,
                451.0902,
                "Deir El-Balah",
                None,
                0,
                0,
                0,
                1204.021,
                "Gaza",
                None,
                0.379586,
                0.1477162,
                0.9718789,
                963.5389,
                "Hebron",
                None,
                0.9127916,
                0.3687683,
                2.241327,
                414.6011,
                "Jenin",
                None,
                0.9302664,
                0.1729868,
                4.841876,
                75.54012,
                "Jericho and Al Aghwar",
                None,
                1.340861,
                0.6852518,
                2.607252,
                479.7003,
                "Jerusalem",
                None,
                0,
                0,
                0,
                645.629,
                "Khan Yunis",
                None,
                0,
                0,
                0,
                450.1119,
                "Nablus",
                None,
                0.1555962,
                0.022686,
                1.058937,
                657.3873,
                "North Gaza",
                None,
                0.7926489,
                0.1494378,
                4.090945,
                140.6539,
                "Qalqiliya",
                None,
                0,
                0,
                0,
                459.5546,
                "Rafah",
                None,
                0.2857052,
                0.040781,
                1.97258,
                331.3463,
                "Ramallah & Al-Bireh",
                None,
                0,
                0,
                0,
                97.44475,
                "Salfit",
                None,
                0,
                0,
                0,
                55.42068,
                "Tubas",
                None,
                0.4782128,
                0.0758192,
                2.953107,
                206.3006,
                "Tulkarm",
            ]

    def test_get_tabular_rows_json(self, fixturejsonurl):
        with Download() as downloader:
            headers, iterator = downloader.get_tabular_rows(
                fixturejsonurl, format="json"
            )
            assert headers == ["Country Code", "Country Name", "Population"]
            assert list(iterator) == [
                ["#country+code", "#country+name", "#population"],
                ["AFG", "Afghanistan", 38041754],
            ]

    def test_download_tabular_rows_as_dicts(self, fixtureprocessurl):
        with Download() as downloader:
            result = downloader.download_tabular_rows_as_dicts(
                fixtureprocessurl, headers=2
            )
            assert result == {
                "coal": {
                    "header2": "3",
                    "header3": "7.4",
                    "header4": "'needed'",
                },
                "gas": {"header2": "2", "header3": "6.5", "header4": "'n/a'"},
            }
            result = downloader.download_tabular_rows_as_dicts(
                fixtureprocessurl, headers=2, keycolumn=2
            )
            assert result == {
                "2": {"header1": "gas", "header3": "6.5", "header4": "'n/a'"},
                "3": {
                    "header1": "coal",
                    "header3": "7.4",
                    "header4": "'needed'",
                },
            }

    def test_download_tabular_cols_as_dicts(self, fixtureprocessurl):
        with Download() as downloader:
            result = downloader.download_tabular_cols_as_dicts(
                fixtureprocessurl, headers=2
            )
            assert result == {
                "header2": {"coal": "3", "gas": "2"},
                "header3": {"coal": "7.4", "gas": "6.5"},
                "header4": {"coal": "'needed'", "gas": "'n/a'"},
            }
            result = downloader.download_tabular_cols_as_dicts(
                fixtureprocessurl, headers=2, keycolumn=2
            )
            assert result == {
                "header1": {"3": "coal", "2": "gas"},
                "header3": {"3": "7.4", "2": "6.5"},
                "header4": {"3": "'needed'", "2": "'n/a'"},
            }

    def test_get_column_positions(self):
        assert Download.get_column_positions(["a", "b", "c"]) == {
            "a": 0,
            "b": 1,
            "c": 2,
        }

    def test_generate_downloaders(self, downloaders):
        downloader1 = Download.get_downloader()
        downloader2 = Download.get_downloader("test")
        assert_downloaders(downloader1, downloader2, downloaders)



"""HTML Tests"""
from os.path import join

import pytest

from hdx.utilities.html import extract_table, get_soup
from hdx.utilities.loader import load_text


class TestHTML:
    url = "https://unstats.un.org/unsd/methodology/m49/overview/"

    @pytest.fixture(scope="function")
    def htmltext(self, fixturesfolder):
        return load_text(join(fixturesfolder, "html", "response.html"))

    @pytest.fixture(scope="function")
    def downloader(self, htmltext):
        class Response:
            pass

        class Download:
            @staticmethod
            def download(url):
                response = Response()
                if url == TestHTML.url:
                    response.text = htmltext
                return response

        return Download()

    def test_html(self, downloader):
        soup = get_soup(TestHTML.url, downloader=downloader)
        tabletag = soup.find(id="downloadTableEN")
        result = extract_table(tabletag)
        assert len(result) == 249
        assert result[10] == {
            "Intermediate Region Name": "Eastern Africa",
            "Global Code": "001",
            "Region Code": "002",
            "Land Locked Developing Countries (LLDC)": "",
            "Intermediate Region Code": "014",
            "Sub-region Name": "Sub-Saharan Africa",
            "Least Developed Countries (LDC)": "x",
            "Global Name": "World",
            "M49 Code": "262",
            "ISO-alpha3 Code": "DJI",
            "Small Island Developing States (SIDS)": "",
            "Developed / Developing Countries": "Developing",
            "Country or Area": "Djibouti",
            "Region Name": "Africa",
            "Sub-region Code": "202",
        }
        assert result[247] == {
            "Global Code": "001",
            "Sub-region Name": "Polynesia",
            "Intermediate Region Code": "",
            "Intermediate Region Name": "",
            "Region Code": "009",
            "Least Developed Countries (LDC)": "x",
            "ISO-alpha3 Code": "TUV",
            "Sub-region Code": "061",
            "Region Name": "Oceania",
            "Developed / Developing Countries": "Developing",
            "Small Island Developing States (SIDS)": "x",
            "Land Locked Developing Countries (LLDC)": "",
            "M49 Code": "798",
            "Global Name": "World",
            "Country or Area": "Tuvalu",
        }



"""Date Parse Utility Tests"""
from datetime import datetime, timedelta, timezone

import pytest
from dateutil.parser import ParserError

from hdx.utilities.dateparse import (
    get_datetime_from_timestamp,
    get_timestamp_from_datetime,
    iso_string_from_datetime,
    now_utc,
    parse_date,
    parse_date_range,
)


class TestDateParse:
    def test_now_utc(self):
        assert now_utc().replace(
            second=0, microsecond=0
        ) == datetime.utcnow().replace(
            second=0, microsecond=0, tzinfo=timezone.utc
        )

    def test_parse_date_range(self):
        result = datetime(2013, 2, 10, 0, 0, tzinfo=timezone.utc), datetime(
            2013, 2, 10, 0, 0, tzinfo=timezone.utc
        )
        assert parse_date_range("10/02/2013") == result
        assert parse_date_range("2013/02/10") == result
        result = datetime(2013, 2, 20, 10, 0, tzinfo=timezone.utc), datetime(
            2013, 2, 20, 10, 0, tzinfo=timezone.utc
        )
        assert parse_date_range("20/02/2013 10:00:00") == result
        assert (
            parse_date_range("20/02/2013 10:00:00", "%d/%m/%Y %H:%M:%S")
            == result
        )
        result = datetime(2013, 2, 20, 0, 0, tzinfo=timezone.utc), datetime(
            2013, 2, 20, 0, 0, tzinfo=timezone.utc
        )
        assert parse_date_range("20/02/2013") == result
        assert (
            parse_date_range("20/02/2013 10:00:00", zero_time=True) == result
        )
        result2 = datetime(2013, 2, 20, 0, 0, tzinfo=timezone.utc), datetime(
            2013, 2, 20, 23, 59, 59, tzinfo=timezone.utc
        )
        assert (
            parse_date_range(
                "20/02/2013 10:00:00",
                zero_time=True,
                max_endtime=True,
            )
            == result2
        )
        result2 = (
            datetime(2013, 2, 20, 23, 59, 59, 999999, tzinfo=timezone.utc),
            datetime(2013, 2, 20, 23, 59, 59, 999999, tzinfo=timezone.utc),
        )
        assert (
            parse_date_range(
                "20/02/2013 10:00:00",
                include_microseconds=True,
                max_starttime=True,
                max_endtime=True,
            )
            == result2
        )
        assert parse_date_range("20/02/2013", "%d/%m/%Y") == result
        assert (
            parse_date_range(
                "20/02/2013 10:00:00", "%d/%m/%Y %H:%M:%S", zero_time=True
            )
            == result
        )
        fuzzy = {}
        assert (
            parse_date_range("date is 20/02/2013 for this test", fuzzy=fuzzy)
            == result
        )
        fuzzyresult = {
            "startdate": datetime(2013, 2, 20, 0, 0, tzinfo=timezone.utc),
            "enddate": datetime(2013, 2, 20, 0, 0, tzinfo=timezone.utc),
            "nondate": ("date is ", " for this test"),
            "date": ("20/02/2013",),
        }
        assert fuzzy == fuzzyresult
        fuzzy = {}
        assert (
            parse_date_range(
                "date is 20/02/2013 for this test",
                date_format="%d/%m/%Y",
                fuzzy=fuzzy,
            )
            == result
        )
        assert fuzzy == fuzzyresult
        fuzzy = {}
        assert (
            parse_date_range("20/02/2013", date_format="%d/%m/%Y", fuzzy=fuzzy)
            == result
        )
        fuzzyresult["nondate"] = None
        assert fuzzy == fuzzyresult
        result = datetime(2013, 2, 1, 0, 0, tzinfo=timezone.utc), datetime(
            2013, 2, 28, 0, 0, tzinfo=timezone.utc
        )
        assert parse_date_range("02/2013") == result
        assert parse_date_range("02/2013", "%m/%Y") == result
        fuzzy = {}
        assert (
            parse_date_range("date is 02/2013 for this test", fuzzy=fuzzy)
            == result
        )
        assert fuzzy == {
            "startdate": datetime(2013, 2, 1, 0, 0, tzinfo=timezone.utc),
            "enddate": datetime(2013, 2, 28, 0, 0, tzinfo=timezone.utc),
            "nondate": ("date is ", " for this test"),
            "date": ("02/2013",),
        }
        result = datetime(2013, 1, 1, 0, 0, tzinfo=timezone.utc), datetime(
            2013, 12, 31, 0, 0, tzinfo=timezone.utc
        )
        assert parse_date_range("2013") == result
        assert parse_date_range("2013", "%Y") == result
        fuzzy = {}
        date = datetime(2001, 12, 10, 0, 0, tzinfo=timezone.utc)
        result = date, date
        assert (
            parse_date_range(
                "State_Village_Tract_Boundaries 10/12/01 lala", fuzzy=fuzzy
            )
            == result
        )
        assert fuzzy == {
            "startdate": date,
            "enddate": date,
            "nondate": ("State_Village_Tract_Boundaries ", " lala"),
            "date": ("10/12/01",),
        }
        with pytest.raises(ParserError):
            parse_date_range("lalala", "%d/%m/%Y")
        with pytest.raises(ParserError):
            parse_date_range("State_Village_Tract_Boundaries", fuzzy=dict())
        with pytest.raises(ParserError):
            fuzzy = {}
            parse_date_range("Mon_State_Village_Tract_Boundaries", fuzzy=fuzzy)
        with pytest.raises(ParserError):
            parse_date_range("20/02")
        with pytest.raises(ParserError):
            parse_date_range("02/20")
        with pytest.raises(ParserError):
            parse_date_range("20/02", "%d/%m")

    def test_parse_date(self):
        assert parse_date("20/02/2013") == datetime(
            2013, 2, 20, 0, 0, tzinfo=timezone.utc
        )
        assert parse_date("20/02/2013", "%d/%m/%Y") == datetime(
            2013, 2, 20, 0, 0, tzinfo=timezone.utc
        )
        assert parse_date("20/02/2013 01:30:20") == datetime(
            2013, 2, 20, 1, 30, 20, tzinfo=timezone.utc
        )
        assert parse_date("20/02/2013 01:30:20 IST") == datetime(
            2013,
            2,
            20,
            1,
            30,
            20,
            tzinfo=timezone.utc,
        )
        assert parse_date(
            "20/02/2013 01:30:20 IST", timezone_handling=1
        ) == datetime(
            2013,
            2,
            20,
            1,
            30,
            20,
        )
        assert parse_date(
            "20/02/2013 01:30:20 IST",
            timezone_handling=2,
        ) == datetime(
            2013,
            2,
            20,
            1,
            30,
            20,
            tzinfo=timezone(timedelta(hours=5, minutes=30)),
        )
        assert parse_date(
            "20/02/2013 01:30:20 IST",
            timezone_handling=3,
        ) == datetime(
            2013,
            2,
            20,
            1,
            30,
            20,
            tzinfo=timezone(timedelta(hours=5, minutes=30)),
        )
        assert parse_date(
            "20/02/2013 01:30:20 IST", timezone_handling=4
        ) == datetime(
            2013,
            2,
            19,
            20,
            00,
            20,
            tzinfo=timezone.utc,
        )
        assert parse_date(
            "20/02/2013 01:30:20 IST", zero_time=True
        ) == datetime(
            2013,
            2,
            20,
            0,
            0,
            0,
            tzinfo=timezone.utc,
        )
        assert parse_date(
            "20/02/2013 01:30:20 IST", max_time=True
        ) == datetime(
            2013,
            2,
            20,
            23,
            59,
            59,
            tzinfo=timezone.utc,
        )
        assert parse_date(
            "20/02/2013 01:30:20 NUT",
            timezone_handling=2,
            default_timezones="-11 X NUT SST",
        ) == datetime(
            2013,
            2,
            20,
            1,
            30,
            20,
            tzinfo=timezone(timedelta(hours=-11)),
        )
        assert parse_date(
            "20/02/2013 01:30:20 IST",
            timezone_handling=2,
            default_timezones="-11 X NUT SST",
        ) == datetime(
            2013,
            2,
            20,
            1,
            30,
            20,
        )
        assert parse_date(
            "20/02/2013 01:30:20 IST",
            timezone_handling=3,
            default_timezones="-11 X NUT SST",
        ) == datetime(
            2013,
            2,
            20,
            1,
            30,
            20,
            tzinfo=timezone.utc,
        )
        assert parse_date(
            "20/02/2013 01:30:20 IST",
            timezone_handling=5,
            default_timezones="-11 X NUT SST",
        ) == datetime(
            2013,
            2,
            20,
            1,
            30,
            20,
            tzinfo=timezone.utc,
        )
        with pytest.raises(ParserError):
            parse_date("02/2013")
        with pytest.raises(ParserError):
            parse_date("02/2013", "%m/%Y")

    def test_get_datetime_from_timestamp(self):
        expected_timestamp = 1596180834.0
        expected_date = datetime(2020, 7, 31, 7, 33, 54, tzinfo=timezone.utc)
        timestamp = get_timestamp_from_datetime(expected_date)
        assert timestamp == expected_timestamp
        date = get_datetime_from_timestamp(
            expected_timestamp, timezone=timezone.utc
        )
        assert date == expected_date
        date = get_datetime_from_timestamp(
            expected_timestamp * 1000, timezone=timezone.utc
        )
        assert date == expected_date

    def test_iso_string_from_datetime(self):
        date = datetime(2020, 7, 31, 7, 33, 54, tzinfo=timezone.utc)
        string = iso_string_from_datetime(date)
        assert string == "2020-07-31"



"""Configuration of logging."""
import logging
import logging.config
from sys import stderr
from typing import Optional

from loguru import logger


def setup_logging(
    console_log_level: str = "INFO",
    log_file: Optional[str] = None,
    file_log_level: str = "ERROR",
) -> None:
    """Setup logging configuration. Intercepts standard logging and outputs
    errors to a file.

    Args:
        console_log_level (str): Log level to use for console output. Defaults to INFO.
        log_file (Optional[str]): Path of log file. Defaults to None (No log file).
        file_log_level (str): Log level to use for console output. Defaults to ERROR.

    Returns:
        None
    """
    logger.remove()
    logger.add(
        stderr,
        level=console_log_level,
        colorize=True,
        backtrace=True,
        diagnose=True,
    )
    if log_file:
        logger.add(
            log_file,
            level=file_log_level,
            mode="w",
            backtrace=True,
            diagnose=True,
        )

    class InterceptHandler(logging.Handler):
        def emit(self, record):
            # Get corresponding Loguru level if it exists
            try:
                level = logger.level(record.levelname).name
            except ValueError:
                level = record.levelno

            # Find caller from where originated the logged message
            frame, depth = logging.currentframe(), 2
            while frame.f_code.co_filename == logging.__file__:
                frame = frame.f_back
                depth += 1

            logger.opt(
                colors=True,
                depth=depth,
                exception=record.exc_info,
            ).log(level, record.getMessage())

    root_logger = logging.getLogger()
    while root_logger.hasHandlers():
        root_logger.removeHandler(root_logger.handlers[0])
    root_logger.setLevel(logging.NOTSET)
    logging.basicConfig(handlers=[InterceptHandler()], level=logging.NOTSET)



"""Utility to save state to a file and read it back."""
import logging
from typing import Any, Callable

from hdx.utilities.loader import load_text
from hdx.utilities.saver import save_text

logger = logging.getLogger(__name__)


class State:
    """State class that allows the reading and writing of state to a given
    path. Input and output state transformations can be supplied in read_fn and
    write_fn respectively. The input state transformation takes in a string
    while the output transformation outputs a string. If run inside a GitHub
    Action, the saved state file could be committed to GitHub so that on next
    run the state is available in the repository.

    Args:
        path (str): Path to save state file
        read_fn (Callable[[str], Any]): Input state transformation. Defaults to lambda x: x.
        write_fn: Callable[[Any], str]: Output state transformation. Defaults to lambda x: x.
    """

    def __init__(
        self,
        path: str,
        read_fn: Callable[[str], Any] = lambda x: x,
        write_fn: Callable[[Any], str] = lambda x: x,
    ) -> None:
        self.path = path
        self.read_fn = read_fn
        self.write_fn = write_fn
        self.state = self.read()

    def __enter__(self) -> "State":
        """Allow usage of with.

        Returns:
            State: SavedState object
        """
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """Allow usage of with.

        Args:
            exc_type (Any): Exception type
            exc_value (Any): Exception value
            traceback (Any): Traceback

        Returns:
            None
        """
        self.write()

    def read(self) -> Any:
        """Read state from file

        Returns:
            None
        """
        value = self.read_fn(load_text(self.path))
        logger.info(f"State read from {self.path} = {value}")
        return value

    def write(self) -> None:
        """Write state to file

        Returns:
            None
        """
        logger.info(f"State written to {self.path} = {self.state}")
        save_text(self.write_fn(self.state), self.path)

    def get(self) -> Any:
        """Get the state

        Returns:
            Any: State
        """
        return self.state

    def set(self, state: Any):
        """Set the state

        Args:
            state (Any): State

        Returns:
            None
        """
        self.state = state



"""Errors on exit Tests"""
import logging

import pytest

from hdx.utilities.easy_logging import setup_logging
from hdx.utilities.errors_onexit import ErrorsOnExit

setup_logging()


class TestErrorsOnExit:
    def test_errorsonexit(self, caplog):
        with ErrorsOnExit() as errors:
            assert len(errors.errors) == 0
        with pytest.raises(SystemExit):
            with caplog.at_level(logging.ERROR):
                with ErrorsOnExit():
                    logging.errors_on_exit.add("error 1")
                    logging.errors_on_exit.add("error 2")
                    assert len(logging.errors_on_exit.errors) == 2
                assert "error 1" in caplog.text
                assert "error 2" in caplog.text



"""Frictionless wrapper"""
from typing import Any, Optional, Tuple

import requests
from frictionless import (
    Control,
    Detector,
    Dialect,
    FrictionlessException,
    Resource,
    system,
)
from frictionless.errors import ResourceError
from frictionless.formats import CsvControl, ExcelControl, JsonControl


def get_frictionless_control(**kwargs: Any) -> Tuple[Control, Any]:
    """Get Frictionless Control.

    Args:
        **kwargs:
        file_type (Optional[str]): Type of file. Defaults to inferring.
        format (Optional[str]): Type of file. Defaults to inferring.
        delimiter (Optional[str]): Delimiter for values in csv rows. Defaults to inferring.
        skip_initial_space (bool): Ignore whitespace straight after delimiter. Defaults to False.
        sheet (Optional[Union[int, str]): Sheet in Excel. Defaults to inferring.
        fill_merged_cells (bool): Whether to fill merged cells. Defaults to True.
        keyed (bool): Whether JSON is keyed. Defaults to True.
        keys (Optional[List[str]]): JSON keys to get. Defaults to None (all of them).
        property (Optional[str]): Path to table in JSON. Defaults to None.
        control (Control): This can be set to override the above. See Frictionless docs.

    Returns:
        Tuple[Control, Any]: (frictionless Control object, kwargs)
    """
    control = kwargs.get("control")
    file_type = kwargs.pop("file_type", None)
    file_format = kwargs.get("format", file_type)
    if file_format is not None:
        kwargs["format"] = file_format
        if control is None:
            if file_format == "csv":
                control = CsvControl()
                delimiter = kwargs.pop("delimiter", None)
                if delimiter is not None:
                    control.delimiter = delimiter
                skip_initial_space = kwargs.pop("skip_initial_space", None)
                if skip_initial_space is not None:
                    control.skip_initial_space = skip_initial_space
            elif file_format in ("xls", "xlsx"):
                control = ExcelControl()
                sheet = kwargs.pop("sheet", None)
                if sheet is not None:
                    control.sheet = sheet
                fill_merged_cells = kwargs.pop("fill_merged_cells", True)
                control.fill_merged_cells = fill_merged_cells
            elif file_format == "json":
                control = JsonControl()
                keyed = kwargs.pop("keyed", True)
                control.keyed = keyed
                keys = kwargs.pop("keys", None)
                if keys is not None:
                    control.keys = keys
                property = kwargs.pop("property", None)
                if property is not None:
                    control.property = property
                kwargs["type"] = "table"
    return control, kwargs


def get_frictionless_detector(
    infer_types: bool, **kwargs: Any
) -> Tuple[Detector, Any]:
    """Get Frictionless Detector.

    Args:
        infer_types (bool): Whether to infer types. Defaults to False (strings).
        **kwargs:
        default_type (Optional[str]): Default field type if infer_types False. Defaults to any.
        float_numbers (bool): Use float not Decimal if infer_types True. Defaults to True.
        null_values (List[Any]): Values that will return None. Defaults to [""].
        detector (Detector): This can be set to override the above. See Frictionless docs.

    Returns:
        Tuple[Detector, Any]: (frictionless Detector object, kwargs)
    """
    detector = kwargs.get("detector", Detector())
    if infer_types:
        default = None
    else:
        default = "any"
    default_type = kwargs.pop("default_type", default)
    detector.field_type = default_type
    float_numbers = kwargs.pop("float_numbers", True)
    detector.field_float_numbers = float_numbers
    null_values = kwargs.pop("null_values", [""])
    detector.field_missing_values = null_values
    return detector, kwargs


def get_frictionless_dialect(
    ignore_blank_rows: bool, **kwargs: Any
) -> Tuple[Dialect, Any]:
    """Get Frictionless Dialect.

    Args:
        ignore_blank_rows (bool): Whether to ignore blank rows. Defaults to True.
        **kwargs:
        columns (Union[ListTuple[int], ListTuple[str], None]): Columns to pick. Defaults to all.
        dialect (Dialect): This can be set to override the above. See Frictionless docs.

    Returns:
        Tuple[Dialect, Any]: (frictionless Dialect object, Any)
    """
    dialect = kwargs.get("dialect", Dialect())
    columns = kwargs.pop("columns", None)
    if columns:
        dialect.pick_fields = columns
    dialect.skip_blank_rows = ignore_blank_rows
    return dialect, kwargs


def get_frictionless_resource(
    url: Optional[str] = None,
    ignore_blank_rows: bool = True,
    infer_types: bool = False,
    session: Optional[requests.Session] = None,
    data: Optional[Any] = None,
    **kwargs: Any,
) -> Resource:
    """Get Frictionless Resource. Either url or data must be supplied.

    Args:
        url (Optional[str]): URL or path to download. Defaults to None.
        ignore_blank_rows (bool): Whether to ignore blank rows. Defaults to True.
        infer_types (bool): Whether to infer types. Defaults to False (strings).
        session (Optional[requests.Session]): Session to use. Defaults to not setting a session.
        data (Optional[Any]): Data to parse. Defaults to None.
        **kwargs:
        has_header (bool): Whether data has a header. Defaults to True.
        headers (Union[int, ListTuple[int], ListTuple[str]]): Number of row(s) containing headers or list of headers.  # pylint: disable=line-too-long
        columns (Union[ListTuple[int], ListTuple[str], None]): Columns to pick. Defaults to all.
        file_type (Optional[str]): Type of file. Defaults to inferring.
        format (Optional[str]): Type of file. Defaults to inferring.
        encoding (Optional[str]): Type of encoding. Defaults to inferring.
        compression (Optional[str]): Type of compression. Defaults to inferring.
        delimiter (Optional[str]): Delimiter for values in csv rows. Defaults to inferring.
        skip_initial_space (bool): Ignore whitespace straight after delimiter. Defaults to False.
        sheet (Optional[Union[int, str]): Sheet in Excel. Defaults to inferring.
        fill_merged_cells (bool): Whether to fill merged cells. Defaults to True.
        keyed (bool): Whether JSON is keyed. Defaults to True.
        keys (Optional[List[str]]): JSON keys to get. Defaults to None (all of them).
        property (Optional[str]): Path to table in JSON. Defaults to None.
        http_session (Session): Session object to use. Defaults to downloader session.
        default_type (Optional[str]): Default field type if infer_types False. Defaults to any.
        float_numbers (bool): Use float not Decimal if infer_types True. Defaults to True.
        null_values (List[Any]): Values that will return None. Defaults to [""].
        control (Control): This can be set to override the above. See Frictionless docs.
        detector (Detector): This can be set to override the above. See Frictionless docs.
        dialect (Dialect): This can be set to override the above. See Frictionless docs.
        schema (Schema): This can be set to override the above. See Frictionless docs.

    Returns:
        Resource: frictionless Resource object
    """
    if not url and not data:
        error = ResourceError(note="Neither url or data supplied!")
        raise FrictionlessException(error=error)
    control, kwargs = get_frictionless_control(**kwargs)
    detector, kwargs = get_frictionless_detector(infer_types, **kwargs)
    dialect, kwargs = get_frictionless_dialect(ignore_blank_rows, **kwargs)
    has_header = kwargs.pop("has_header", None)
    headers = kwargs.pop("headers", None)
    if headers is not None:
        if isinstance(headers, int):
            headers = [headers]
        if isinstance(headers[0], int):
            dialect.header_rows = headers
        else:
            detector.field_names = headers
            if has_header is None:
                has_header = False
    if has_header is None:
        has_header = True
    dialect.header = has_header
    kwargs["detector"] = detector
    kwargs["dialect"] = dialect
    if control:
        dialect.add_control(control)
    http_session = kwargs.pop("http_session", session)
    with system.use_context(http_session=http_session):
        if url:
            resource = Resource(path=url, **kwargs)
        else:
            resource = Resource(data=data, **kwargs)
        resource.open()
        return resource



"""Saver Tests"""
import json
from collections import OrderedDict
from os.path import exists, join

import pytest

from hdx.utilities.compare import assert_files_same
from hdx.utilities.loader import load_yaml
from hdx.utilities.saver import save_hxlated_output, save_json, save_yaml


class TestLoader:
    yaml_to_write = OrderedDict(
        [
            (
                "hdx_prod_site",
                OrderedDict(
                    [
                        ("url", "https://data.humdata.org/"),
                        ("username", None),
                        ("password", None),
                    ]
                ),
            ),
            (
                "hdx_test_site",
                OrderedDict(
                    [
                        ("url", "https://test-data.humdata.org/"),
                        ("username", "lala"),
                        ("password", "lalala"),
                    ]
                ),
            ),
            (
                "dataset",
                OrderedDict(
                    [("required_fields", ["name", "title", "dataset_date"])]
                ),
            ),
            (
                "resource",
                OrderedDict(
                    [
                        (
                            "required_fields",
                            ["package_id", "name", "description"],
                        )
                    ]
                ),
            ),
            (
                "showcase",
                OrderedDict([("required_fields", ["name", "title"])]),
            ),
            ("param_1", "ABC"),
        ]
    )

    json_to_write = OrderedDict(
        [
            (
                "hdx_prod_site",
                OrderedDict(
                    [
                        ("url", "https://data.humdata.org/"),
                        ("username", None),
                        ("password", None),
                    ]
                ),
            ),
            (
                "hdx_test_site",
                OrderedDict(
                    [
                        ("url", "https://test-data.humdata.org/"),
                        ("username", "tumteetum"),
                        ("password", "tumteetumteetum"),
                    ]
                ),
            ),
            (
                "dataset",
                OrderedDict([("required_fields", ["name", "dataset_date"])]),
            ),
            (
                "resource",
                OrderedDict([("required_fields", ["name", "description"])]),
            ),
            ("showcase", OrderedDict([("required_fields", ["name"])])),
            ("my_param", "abc"),
        ]
    )

    @pytest.fixture(scope="class")
    def saverfolder(self, fixturesfolder):
        return join(fixturesfolder, "saver")

    @pytest.fixture(scope="class")
    def json_csv_configuration(self, fixturesfolder):
        return load_yaml(join(fixturesfolder, "config", "json_csv.yml"))

    @pytest.mark.parametrize(
        "filename,pretty,sortkeys",
        [
            ("pretty-false_sortkeys-false.yaml", False, False),
            ("pretty-false_sortkeys-true.yaml", False, True),
            ("pretty-true_sortkeys-false.yaml", True, False),
            ("pretty-true_sortkeys-true.yaml", True, True),
        ],
    )
    def test_save_yaml(self, tmpdir, saverfolder, filename, pretty, sortkeys):
        test_path = join(str(tmpdir), filename)
        ref_path = join(saverfolder, filename)
        save_yaml(
            TestLoader.yaml_to_write,
            test_path,
            pretty=pretty,
            sortkeys=sortkeys,
        )
        assert_files_same(ref_path, test_path)
        dct = json.loads(json.dumps(TestLoader.yaml_to_write))
        save_yaml(dct, test_path, pretty=pretty, sortkeys=sortkeys)
        assert_files_same(ref_path, test_path)

    @pytest.mark.parametrize(
        "filename,pretty,sortkeys",
        [
            ("pretty-false_sortkeys-false.json", False, False),
            ("pretty-false_sortkeys-true.json", False, True),
            ("pretty-true_sortkeys-false.json", True, False),
            ("pretty-true_sortkeys-true.json", True, True),
        ],
    )
    def test_save_json(self, tmpdir, saverfolder, filename, pretty, sortkeys):
        test_path = join(str(tmpdir), filename)
        ref_path = join(saverfolder, filename)
        save_json(
            TestLoader.json_to_write,
            test_path,
            pretty=pretty,
            sortkeys=sortkeys,
        )
        assert_files_same(ref_path, test_path)
        dct = json.loads(json.dumps(TestLoader.json_to_write))
        save_json(dct, test_path, pretty=pretty, sortkeys=sortkeys)
        assert_files_same(ref_path, test_path)

    def test_save_hxlated_output(
        self, tmpdir, saverfolder, json_csv_configuration
    ):
        rows = (
            ("Col1", "Col2", "Col3"),
            ("#tag1", "#tag2", "#tag3"),
            (1, "2", 3),
            (4, "5", 6),
        )
        output_dir = str(tmpdir)

        save_hxlated_output(
            json_csv_configuration["test1"],
            rows[2:],
            includes_header=False,
            includes_hxltags=False,
            output_dir=output_dir,
        )
        filename = "out.csv"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )
        filename = "out.json"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )

        row0 = rows[0]
        rowsdict = []
        for row in rows[2:]:
            newrow = {}
            for i, key in enumerate(row0):
                newrow[key] = row[i]
            rowsdict.append(newrow)
        save_hxlated_output(
            json_csv_configuration["test1"],
            rowsdict,
            includes_header=False,
            includes_hxltags=False,
            output_dir=output_dir,
        )
        filename = "out.csv"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )
        filename = "out.json"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )

        save_hxlated_output(
            json_csv_configuration["test2"],
            rows,
            includes_header=True,
            includes_hxltags=True,
            output_dir=output_dir,
        )
        filename = "out2.csv"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )
        filename = "out2.json"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )

        rowsdict = []
        for row in rows[1:]:
            newrow = {}
            for i, key in enumerate(row0):
                newrow[key] = row[i]
            rowsdict.append(newrow)
        save_hxlated_output(
            json_csv_configuration["test2"],
            rowsdict,
            includes_header=True,
            includes_hxltags=True,
            output_dir=output_dir,
        )
        filename = "out2.csv"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )
        filename = "out2.json"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )

        save_hxlated_output(
            json_csv_configuration["test3"],
            rows,
            includes_header=True,
            includes_hxltags=True,
            output_dir=output_dir,
        )
        filename = "out3.csv"
        assert_files_same(
            join(saverfolder, "out.csv"), join(output_dir, filename)
        )
        filename = "out3.json"
        assert exists(join(output_dir, filename)) is False

        save_hxlated_output(
            json_csv_configuration["test4"],
            rows,
            includes_header=True,
            includes_hxltags=True,
            output_dir=output_dir,
        )
        filename = "out4.csv"
        assert exists(join(output_dir, filename)) is False
        filename = "out4.json"
        assert_files_same(
            join(saverfolder, "out2.json"), join(output_dir, filename)
        )

        save_hxlated_output(
            json_csv_configuration["test5"],
            rowsdict,
            includes_header=True,
            includes_hxltags=True,
            output_dir=output_dir,
        )
        filename = "out5.csv"
        assert_files_same(
            join(saverfolder, "out2.csv"), join(output_dir, filename)
        )
        filename = "out5.json"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )

        save_hxlated_output(
            json_csv_configuration["test6"],
            rowsdict,
            includes_header=True,
            includes_hxltags=True,
            output_dir=output_dir,
            today="today!",
        )
        filename = "out6.csv"
        assert_files_same(
            join(saverfolder, "out2.csv"), join(output_dir, filename)
        )
        filename = "out6.json"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )

        save_hxlated_output(
            json_csv_configuration["test7"],
            rowsdict,
            includes_header=True,
            includes_hxltags=True,
            output_dir=output_dir,
            today="today!",
        )
        filename = "out7.csv"
        assert_files_same(
            join(saverfolder, "out2.csv"), join(output_dir, filename)
        )
        filename = "out7.json"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )

        save_hxlated_output(
            json_csv_configuration["test8"],
            rowsdict,
            includes_header=True,
            includes_hxltags=True,
            output_dir=output_dir,
            today="today!",
        )
        filename = "out8.csv"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )
        filename = "out8.json"
        assert_files_same(
            join(saverfolder, filename), join(output_dir, filename)
        )



"""Encoding Utility Tests"""
from hdx.utilities.encoding import (
    base64_to_str,
    basicauth_decode,
    basicauth_encode,
    str_to_base64,
)


class TestEncoding:
    def test_base64(self):
        string = "hello"
        result = str_to_base64(string)
        assert result == "aGVsbG8="
        assert base64_to_str(result) == string

    def test_basicauth(self):
        user = "user"
        password = "password"
        result = basicauth_encode(user, password)
        assert result == "Basic dXNlcjpwYXNzd29yZA=="
        result = basicauth_decode(result)
        assert result == (user, password)



def assert_downloaders(downloader1, downloader2, params):
    user_agent, custom_user_agent, extra_params_dict = params
    test_url = "http://www.lalala.com/lala"
    assert downloader1.session.headers["User-Agent"].endswith(user_agent)
    assert downloader1.session.auth is None
    assert downloader1.get_full_url(test_url) == test_url
    assert downloader2.session.headers["User-Agent"].endswith(
        custom_user_agent
    )
    assert downloader2.session.auth == ("user", "pass")
    key, value = next(iter(extra_params_dict.items()))
    assert downloader2.get_full_url(test_url) == f"{test_url}?{key}={value}"



"""File compare utilities."""
import difflib
from os import linesep
from typing import List


def compare_files(path1: str, path2: str) -> List[str]:
    """Returns the delta between two files using -, ?, + format excluding lines
    that are the same.

    Args:
        path1 (str): Path to first file
        path2 (str): Path to second file

    Returns:
        List[str]: Delta between the two files
    """
    diff = difflib.ndiff(
        open(path1).read().splitlines(), open(path2).read().splitlines()
    )
    return [x for x in diff if x[0] in ["-", "+", "?"]]


def assert_files_same(path1: str, path2: str) -> None:
    """Asserts that two files are the same and returns delta using.

    -, ?, + format if not

    Args:
        path1 (str): Path to first file
        path2 (str): Path to second file

    Returns:
        None
    """
    difflines = compare_files(path1, path2)
    assert len(difflines) == 0, linesep.join([linesep] + difflines)



"""Utility class to simplify sending emails."""

import logging
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from os.path import expanduser, join
from typing import Any, List, Optional, Union

from .loader import load_json, load_yaml
from .typehint import ListTuple

try:
    from email_validator import validate_email
except ImportError:
    validate_email = None

logger = logging.getLogger(__name__)


class EmailConfigurationError(Exception):
    pass


class Email:
    """Emailer utility. Parameters in dictionary or file (eg. yaml below):

     | connection_type: "ssl"   ("ssl" for smtp ssl or "lmtp", otherwise basic smtp is assumed)
     | host: "localhost"
     | port: 123
     | local_hostname: "mycomputer.fqdn.com"
     | timeout: 3
     | username: "user"
     | password: "pass"
     | sender: "a@b.com"    (if not supplied username is used as sender)

    Args:
        **kwargs: See below
        email_config_dict (dict): HDX configuration dictionary OR
        email_config_json (str): Path to JSON HDX configuration OR
        email_config_yaml (str): Path to YAML HDX configuration. Defaults to ~/hdx_email_configuration.yml.
    """

    default_email_config_yaml = join(
        expanduser("~"), "hdx_email_configuration.yml"
    )

    def __init__(self, **kwargs: Any) -> None:
        email_config_found = False
        email_config_dict = kwargs.get("email_config_dict", None)
        if email_config_dict:
            email_config_found = True
            logger.info("Loading email configuration from dictionary")

        email_config_json = kwargs.get("email_config_json", "")
        if email_config_json:
            if email_config_found:
                raise EmailConfigurationError(
                    "More than one email configuration file given!"
                )
            email_config_found = True
            logger.info(
                f"Loading email configuration from: {email_config_json}"
            )
            email_config_dict = load_json(email_config_json)

        email_config_yaml = kwargs.get("email_config_yaml", None)
        if email_config_found:
            if email_config_yaml:
                raise EmailConfigurationError(
                    "More than one email configuration file given!"
                )
        else:
            if not email_config_yaml:
                logger.info(
                    "No email configuration parameter. Using default email configuration path."
                )
                email_config_yaml = Email.default_email_config_yaml
            logger.info(
                f"Loading email configuration from: {email_config_yaml}"
            )
            email_config_dict = load_yaml(email_config_yaml)

        self.connection_type = email_config_dict.get("connection_type", "smtp")
        self.host = email_config_dict.get("host", "")
        self.port = email_config_dict.get("port", 0)
        self.local_hostname = email_config_dict.get("local_hostname")
        self.timeout = email_config_dict.get("timeout")
        self.source_address = email_config_dict.get("source_address")
        self.username = email_config_dict.get("username")
        self.password = email_config_dict.get("password")
        self.sender = email_config_dict.get("sender", self.username)
        self.server = None

    def __enter__(self) -> "Email":
        """Return Email object for with statement.

        Returns:
            None
        """
        return self

    def __exit__(self, *args: Any) -> None:
        """Close Email object for end of with statement.

        Args:
            *args: Not used

        Returns:
            None
        """

    def connect(self) -> None:
        """Connect to server.

        Returns:
            None
        """
        if self.connection_type.lower() == "ssl":
            self.server = smtplib.SMTP_SSL(
                host=self.host,
                port=self.port,
                local_hostname=self.local_hostname,
                timeout=self.timeout,
                source_address=self.source_address,
            )
        elif self.connection_type.lower() == "lmtp":
            self.server = smtplib.LMTP(
                host=self.host,
                port=self.port,
                local_hostname=self.local_hostname,
                source_address=self.source_address,
            )
        else:
            self.server = smtplib.SMTP(
                host=self.host,
                port=self.port,
                local_hostname=self.local_hostname,
                timeout=self.timeout,
                source_address=self.source_address,
            )
        self.server.login(self.username, self.password)

    def close(self) -> None:
        """Close connection to email server.

        Returns:
            None
        """
        self.server.quit()

    @staticmethod
    def get_normalised_emails(
        recipients: Union[str, ListTuple[str]]
    ) -> List[str]:
        """Get list of normalised emails.

        Args:
            recipients (Union[str, ListTuple[str]]): Email recipient(s)

        Returns:
            List[str]: Normalised emails
        """
        if isinstance(recipients, str):
            recipients = (recipients,)
        if validate_email is None:
            return recipients
        normalised_recipients = []
        for recipient in recipients:
            v = validate_email(
                recipient, check_deliverability=True
            )  # validate and get info
            normalised_recipients.append(
                v["email"]
            )  # replace with normalized form
        return normalised_recipients

    def send(
        self,
        to: Union[str, ListTuple[str]],
        subject: str,
        text_body: str,
        html_body: Optional[str] = None,
        sender: Optional[str] = None,
        cc: Union[str, ListTuple[str], None] = None,
        bcc: Union[str, ListTuple[str], None] = None,
        **kwargs: Any,
    ) -> None:
        """Send email. to, cc and bcc take either a string email address or a
        list of string email addresses. cc and bcc default to None.

        Args:
            to (Union[str, ListTuple[str]]): Email recipient(s)
            subject (str): Email subject
            text_body (str): Plain text email body
            html_body (Optional[str]): HTML email body
            sender (Optional[str]): Email sender. Defaults to global sender.
            cc (Union[str, ListTuple[str], None]): Email cc. Defaults to None.
            bcc (Union[str, ListTuple[str], None]): Email bcc. Defaults to None.
            **kwargs: See below
            mail_options (List): Mail options (see smtplib documentation)
            rcpt_options (List): Recipient options (see smtplib documentation)

        Returns:
            None
        """
        if sender is None:
            sender = self.sender
        v = validate_email(
            sender, check_deliverability=False
        )  # validate and get info
        sender = v["email"]  # replace with normalized form

        if html_body is not None:
            msg = MIMEMultipart("alternative")
            part1 = MIMEText(text_body, "plain")
            part2 = MIMEText(html_body, "html")
            msg.attach(part1)
            msg.attach(part2)
        else:
            msg = MIMEText(text_body)
        msg["Subject"] = subject
        msg["From"] = sender

        normalised_to = self.get_normalised_emails(to)
        msg["To"] = ", ".join(normalised_to)

        if cc is not None:
            normalised_cc = self.get_normalised_emails(cc)
            msg["Cc"] = ", ".join(normalised_cc)
            normalised_to.extend(normalised_cc)

        if bcc is not None:
            normalised_bcc = self.get_normalised_emails(bcc)
            normalised_to.extend(normalised_bcc)

        # Perform operations via server
        self.connect()
        self.server.sendmail(sender, normalised_to, msg.as_string(), **kwargs)
        self.close()



"""Text processing utilities."""
import difflib
import logging
import re
import string
from string import punctuation
from typing import Any, Dict, List, Optional, Set, Tuple

from hdx.utilities.typehint import ListTuple

logger = logging.getLogger(__name__)


PUNCTUATION_MINUS_BRACKETS = r"""!"#$%&'*+,-./:;<=>?@\^_`|~"""
TEMPLATE_VARIABLES = re.compile("{{.*?}}")


def remove_end_characters(
    string: str, characters_to_remove: str = punctuation
) -> str:
    """Remove any characters at end of string that are in characters_to_remove.

    Args:
        string (str): Input string
        characters_to_remove (str): Characters to remove. Defaults to punctuation.

    Returns:
        str: String with any characters at end of string that are in characters_to_remove removed
    """
    while string[-1] in characters_to_remove:
        string = string[:-1]
    return string


def remove_from_end(
    string: str,
    things_to_remove: List[str],
    logging_text: Optional[str] = None,
    whole_words: bool = True,
) -> str:
    """Remove list of items from end of string, stripping any whitespace.

    Args:
        string (str): Input string
        things_to_remove (List[str]): Things to remove from the end of string
        logging_text (Optional[str]): Text to log. Defaults to None.
        whole_words (bool): Remove parts of or whole words. Defaults to True (whole words only).

    Returns:
        str: String with text removed
    """
    for thing in things_to_remove:
        thing_len = len(thing)
        string_len = len(string)
        if string_len <= thing_len + 1:
            continue
        position = -thing_len
        if string[position:] != thing:
            continue
        if whole_words and string[position - 1].isalpha():
            continue
        newstring = string[:position].strip()
        if logging_text:
            logger.info(logging_text % (string, newstring))
        string = newstring
    return string


def remove_string(
    string: str, toremove: str, end_characters_to_remove: str = punctuation
) -> str:
    """
    Remove string from another string and delete any preceding end characters - by default punctuation (eg. comma)
    and any whitespace following the punctuation

    Args:
        string (str): String to process
        toremove (str): String to remove
        end_characters_to_remove (str): Characters to remove. Defaults to punctuation.

    Returns:
        str: String with other string removed

    """
    index = string.find(toremove)
    newstring = string[:index].strip()
    newstring = remove_end_characters(
        newstring, characters_to_remove=end_characters_to_remove
    )
    return f"{newstring}{string[index + len(toremove):]}"


def multiple_replace(string: str, replacements: Dict[str, str]) -> str:
    """Simultaneously replace multiple strings in a string.

    Args:
        string (str): Input string
        replacements (Dict[str,str]): Replacements dictionary

    Returns:
        str: String with replacements
    """
    if not replacements:
        return string
    pattern = re.compile(
        "|".join(
            [re.escape(k) for k in sorted(replacements, key=len, reverse=True)]
        ),
        flags=re.DOTALL,
    )
    return pattern.sub(lambda x: replacements[x.group(0)], string)


def get_words_in_sentence(sentence: str) -> List[str]:
    """Returns list of words in a sentence.

    Args:
        sentence (str): Sentence

    Returns:
        List[str]: List of words in sentence
    """
    return re.sub(
        "[" + punctuation.replace("'", "") + "]", " ", sentence
    ).split()


def get_matching_text_in_strs(
    a: str,
    b: str,
    match_min_size: int = 30,
    ignore: str = "",
    end_characters: str = "",
) -> List[str]:
    """Returns a list of matching blocks of text in a and b.

    Args:
        a (str): First string to match
        b (str): Second string to match
        match_min_size (int): Minimum block size to match on. Defaults to 30.
        ignore (str): Any characters to ignore in matching. Defaults to ''.
        end_characters (str): End characters to look for. Defaults to ''.

    Returns:
        List[str]: List of matching blocks of text
    """
    compare = difflib.SequenceMatcher(lambda x: x in ignore)
    compare.set_seqs(a=a, b=b)
    matching_text = []

    for match in compare.get_matching_blocks():
        start = match.a
        text = a[start : start + match.size]
        if end_characters:
            prev_text = text
            while len(text) != 0 and text[0] in end_characters:
                text = text[1:]
            while len(text) != 0 and text[-1] not in end_characters:
                text = text[:-1]
            if len(text) == 0:
                text = prev_text
        if len(text) >= match_min_size:
            matching_text.append(text)
    return matching_text


def get_matching_text(
    string_list: List[str],
    match_min_size: int = 30,
    ignore: str = "",
    end_characters: str = ".!\r\n",
) -> str:
    """Returns a string containing matching blocks of text in a list of strings
    followed by non-matching.

    Args:
        string_list (List[str]): List of strings to match
        match_min_size (int): Minimum block size to match on. Defaults to 30.
        ignore (str): Any characters to ignore in matching. Defaults to ''.
        end_characters (str): End characters to look for. Defaults to '.\r\n'.

    Returns:
        str: String containing matching blocks of text followed by non-matching
    """
    a = string_list[0]
    for i in range(1, len(string_list)):
        b = string_list[i]
        result = get_matching_text_in_strs(
            a,
            b,
            match_min_size=match_min_size,
            ignore=ignore,
            end_characters=end_characters,
        )
        a = "".join(result)
    return a


def get_matching_then_nonmatching_text(
    string_list: List[str],
    separator: str = "",
    match_min_size: int = 30,
    ignore: str = "",
    end_characters: str = ".!\r\n",
) -> str:
    """Returns a string containing matching blocks of text in a list of strings
    followed by non-matching.

    Args:
        string_list (List[str]): List of strings to match
        separator (str): Separator to add between blocks of text. Defaults to ''.
        match_min_size (int): Minimum block size to match on. Defaults to 30.
        ignore (str): Any characters to ignore in matching. Defaults to ''.
        end_characters (str): End characters to look for. Defaults to '.\r\n'.

    Returns:
        str: String containing matching blocks of text followed by non-matching
    """

    def add_separator_if_needed(text_list):
        if (
            separator
            and len(text_list) > 0
            and text_list[-1][-len(separator) :] != separator
        ):
            text_list.append(separator)

    a = string_list[0]
    for i in range(1, len(string_list)):
        b = string_list[i]
        combined_len = len(a) + len(b)
        result = get_matching_text_in_strs(
            a,
            b,
            match_min_size=match_min_size,
            ignore=ignore,
            end_characters=end_characters,
        )
        new_a = a
        new_b = b
        for text in result:
            new_a = new_a.replace(text, "")
            new_b = new_b.replace(text, "")
        if new_a and new_a in a:
            pos_a = a.index(new_a)
        else:
            pos_a = combined_len
        if new_b and new_b in b:
            pos_b = b.index(new_b)
        else:
            pos_b = combined_len
        if pos_b > pos_a:
            text_1 = new_b
            pos_1 = pos_b
            text_2 = new_a
            pos_2 = pos_a
        else:
            text_1 = new_a
            pos_1 = pos_a
            text_2 = new_b
            pos_2 = pos_b
        output = []
        pos = 0
        for text in result:
            output.append(text)
            pos += len(text)
            if text_1 and pos >= pos_1:
                add_separator_if_needed(output)
                output.append(text_1)
                pos += len(text_1)
                text_1 = None
            if text_2 and pos >= pos_2:
                add_separator_if_needed(output)
                output.append(text_2)
                pos += len(text_2)
                text_2 = None
        if text_1 and pos_1 == combined_len:
            add_separator_if_needed(output)
            output.append(text_1)
        if text_2 and pos_2 == combined_len:
            add_separator_if_needed(output)
            output.append(text_2)
        a = "".join(output)
    return a


def number_format(
    val: Any, format: str = "%.4f", trailing_zeros: bool = True
) -> str:
    """Format float-castable input as string.

    Args:
        val (float): Number to format
        format (str): Format to use. Defaults to %.4f.
        trailing_zeros (bool): Leave trailing zeros. Defaults to True.

    Returns:
        str: Formatted number as string
    """
    if val == "" or val is None:
        return ""
    val = format % float(val)
    if trailing_zeros:
        return val
    return val.rstrip("0").rstrip(".")


def get_fraction_str(
    numerator: Any,
    denominator: Optional[Any] = None,
    format: str = "%.4f",
    trailing_zeros: bool = True,
) -> str:
    """Given float-castable numerator and optional float-castable denominator,
    format as string, returning '' for invalid numerator or 0 denominator.

    Args:
        numerator (float): Numerator
        denominator (Optional[float]): Denominator. Defaults to None.
        format (str): Format to use. Defaults to %.4f.
        trailing_zeros (bool): Leave trailing zeros. Defaults to True.

    Returns:
        str: Formatted number as string
    """
    try:
        numerator = float(numerator)
        if denominator:
            numerator /= float(denominator)
        else:
            if denominator is not None:
                return ""
        return number_format(numerator, format, trailing_zeros)
    except ValueError:
        pass
    return ""


def only_allowed_in_str(test_str: str, allowed_chars: Set) -> bool:
    """Returns True if test string contains only allowed characters, False if
    not.

    Args:
        test_str (str): Test string
        allowed_chars (Set): Set of allowed characters

    Returns:
        bool: True if test string contains only allowed characters, False if not
    """
    return set(test_str) <= allowed_chars


allowed_numeric = set(string.digits + "." + "," + "%" + "-")


def get_numeric_if_possible(value: Any) -> Any:
    """Return val if it is not a string, otherwise see if it can be cast to
    float or int, taking into account commas and periods.

    Args:
        value (Any): Value

    Returns:
        Any: Value
    """

    def get_int_value(val, denominator):
        val = int(val)
        if denominator != 1:
            return float(val) / denominator
        else:
            return val

    if isinstance(value, str):
        val = value.strip()
        if val != "" and only_allowed_in_str(val, allowed_numeric):
            try:
                minusindex = val.index("-")
                if minusindex != 0:
                    return val
            except ValueError:
                pass
            try:
                commaindex = val.index(",")
            except ValueError:
                commaindex = None
            try:
                dotindex = val.index(".")
            except ValueError:
                dotindex = None
            if val[-1] == "%":
                denominator = 100
                val = val[:-1]
            else:
                denominator = 1
            if commaindex is None:
                if dotindex is None:
                    return get_int_value(val, denominator)
                else:
                    if val.count(".") == 1:
                        return float(val) / denominator
                    else:
                        return get_int_value(val.replace(".", ""), denominator)
            else:
                if dotindex is None:
                    return get_int_value(val.replace(",", ""), denominator)
                else:
                    if dotindex > commaindex:
                        val = val.replace(",", "")
                    else:
                        val = val.replace(".", "")
                        val = val.replace(",", ".")
                    return float(val) / denominator
    return value


def earliest_index(
    string_to_search: str, strings_to_try: ListTuple[str]
) -> Optional[int]:
    """Search a string for each of a list of strings and return the earliest
    index.

    Args:
        string_to_search (str): String to search
        strings_to_try (ListTuple[str]): Strings to try

    Returns:
        Optional[int]: Earliest index of the strings to try in string to search or None
    """
    after_string = len(string_to_search) + 1
    indices = []
    for string_to_try in strings_to_try:
        try:
            index = string_to_search.index(string_to_try)
            indices.append(index)
        except ValueError:
            indices.append(after_string)
    earliest_index = sorted(indices)[0]
    if earliest_index == after_string:
        return None
    else:
        return earliest_index


def match_template_variables(
    string: str,
) -> Tuple[Optional[str], Optional[str]]:
    """Try to match {{XXX}} in input string.

    Args:
        string (str): String in which to look for template

    Returns:
        Tuple[Optional[str], Optional[str]]: (Matched string with brackets, matched string without brackets)
    """
    match = TEMPLATE_VARIABLES.search(string)
    if match:
        template_string = match.group()
        return template_string, template_string[2:-2]
    return None, None
